!   EFTCAMB_Oct14
!   Main EFTCAMB file.
!
!   This file contains all the Dark Energy/Modified Gravity physics
!   that is later used by the modified version of CAMB to compute
!   cosmological observables in the Dark Energy/Modified Graviy scenario
!   that the user chooses.
!   For more informations about the subroutine contained in this file see
!   the numerical notes: arXiv:1405.3590
!

! -------------------------------------------------------------------------------------------------

!Definitions of the relevant EFT technical flags and physical constants later used by the code.
module EFTDef
    use Precision
    use ModelParams
    implicit none
    ! EFT technical flags:
    ! 1) Turn on pi field flag:
    !    Sets the scale factor at which the code starts to evolve the pi field.
    !    At times earlier than these the code evolves perturbations as in GR.
    !    This number is used as a lower bound and is refined if the teory is very close to GR
    !    by the EFTreturntoGR module.
    real(dl), parameter :: EFTturnonpiInitial = 1.d-2
    ! 2) Effective turn on pi field flag:
    !    This is extablished by the code as the time at which the model considered
    !    is so close to GR that there is no seizable DE/MG effect. If such a time exists then
    !    it will override the previous value otherwise it will be equal to that value.
    real(dl) :: EFTturnonpi
    ! 3) Return to GR flag:
    !    This is the bound at which a theory is considered to be exactly GR.
    real(dl), parameter :: EFTtoGR = 1.d-8
    ! 4) Early time stability:
    !    EFTCAMB checks the stability of the theory at all times making sure that the choosen model
    !    is viable during radiation, matter and DE eras.
    !    It is possible to enforce stability only at late times (from EFTturnonpiInitial to today)
    !    by setting this flag to false.
    !    This choice will however make the results dependent on what one calls late time,
    !    i.e. the choice of EFTturnonpiInitial, so we advice to quote it clearly when reporting results.
    logical, parameter :: EarlyTimeStability = .true.
    ! 5) Priors flags:
    !    These flag will decide wether to impose a prior or not.
    !       1- Speed of light prior: the propagation speed of pi should be smaller than the speed of light.
    logical, parameter :: EFTlightspeedPrior = .false.
    !       2- Effective pi mass prior: the effective mass of pi should be greater than zero.
    !          Warning: disabling this could result in numerical instabilities.
    logical, parameter :: EFTpiMassPrior = .true.
    !       3- Additional priors on cosmological parameters. For more informations read the related section
    !          in the numerical notes.
    logical, parameter :: EFTAdditionalPriors = .true.
    ! 6) Use with cosmomc flag:
    !    If EFTCAMB is used inside CosmoMC then the initialization subroutine has to be called
    !    inside it so we don't need to call it inside the CAMB.
    !    This flag takes care of it. Do not change it, as the
    !    CosmoMC code will take care of handeling it.
    integer :: EFTCAMBuseinCOSMOMC = 1
    ! 7) EFTCAMB debug flags:
    !    These flags are used to deeply debug EFTCAMB.
    !    Don't turn them on unless you know what they are going to do.
    !    These flags have to be used in conjunction with two other changes.
    !           1- The fixq flag in cmbmain should be set different from zero.
    !           2- The path of the files created within cmbmain should be consistent
    !              (search EFTCAMB mod in cmbmain).
    !    To debug the code it should be launched on just one thread by
    !    changing the appropriate entry in the parameter file.
    logical, parameter :: DebugEFT       = .false.
    logical, parameter :: DebugEFTOutput = .false.
    logical, parameter :: DebugDesigner  = .false.
    logical, parameter :: DebugMapping   = .false.
    !  EFT cosmological and physical constants:
    ! 1) Physical constants:
    real(dl), parameter :: const_pi_EFT = 3.1415926535897932384626433832795_dl
    real(dl), parameter :: c_EFT = 2.99792458d8
    real(dl), parameter :: G_EFT =6.67428e-11_dl
    real(dl), parameter :: Mpc_EFT = 3.085678e22_dl
    real(dl), parameter :: sigma_boltz_EFT = 5.6704e-8_dl
    real(dl), parameter :: kappa_EFT =8._dl*const_pi_EFT*G_EFT

end module EFTDef

! -------------------------------------------------------------------------------------------------

!Definition of the dark energy EOS used by the designer and pure EFT code.
module EFTdeEOS
    use EFTDef
    implicit none
contains

    !DEC$ ATTRIBUTES FORCEINLINE :: EFTw
    function EFTw(a,deriv)
        implicit none
        real(dl), intent(in) :: a
        integer , intent(in) :: deriv
        real(dl) :: EFTw, a0
        !Protection against numerical errors at very early times
        if (a .lt. 1.d-8) then
            a0 = 1.d-8
        else
            a0 = a
        end if
        ! Definition of the function itself
        !   The user has to define:
        !   deriv= 0: the function w(a) itself
        !   deriv= 1: the first derivative dw(a)/da wrt the scale factor
        !   deriv= 2: the second derivative d^2w(a)/da^2 wrt the scale factor
        !   deriv= 3: the integral (see the numerical notes).
        select case (CP%EFTwDE)
            case (0) ! LCDM
                if (deriv==0) then
                    EFTw = -1._dl
                else if (deriv==1) then
                    EFTw = 0._dl
                else if (deriv==2) then
                    EFTw = 0._dl
                else if (deriv==3) then
                    EFTw = a0*a0
                end if
            case (1) ! wCDM
                if (deriv==0) then
                    EFTw = CP%EFTw0
                else if (deriv==1) then
                    EFTw = 0._dl
                else if (deriv==2) then
                    EFTw = 0._dl
                else if (deriv==3) then
                    EFTw = a0**(-1._dl-3._dl*CP%EFTw0)
                end if
            case (2) ! CPL
                if (deriv==0) then
                    EFTw = CP%EFTw0 + CP%EFTwa*(1._dl-a0)
                else if (deriv==1) then
                    EFTw = -CP%EFTwa
                else if (deriv==2) then
                    EFTw = 0._dl
                else if (deriv==3) then
                    EFTw = a0**(-1._dl-3._dl*(+CP%EFTw0+CP%EFTwa))*exp(3._dl*CP%EFTwa*(a0-1._dl))
                end if
            case (3) ! JPB
                if (deriv==0) then
                    EFTw = CP%EFTw0 +CP%EFTwa*(1._dl-a0)*a0**(CP%EFTwn-1._dl)
                else if (deriv==1) then
                    EFTw = -CP%EFTwa*a0**(CP%EFTwn-2._dl)*(1._dl+(a0-1._dl)*CP%EFTwn)
                else if (deriv==2) then
                    EFTw = -CP%EFTwa*a0**(CP%EFTwn-3._dl)*(CP%EFTwn-1._dl)*(2._dl+(a0-1._dl)*CP%EFTwn)
                else if (deriv==3) then
                    if (CP%EFTwn/=1) then
                        EFTw = a0**(-1._dl-3._dl*CP%EFTw0)*exp((3._dl*(a0+a0**CP%EFTwn*(a0*(CP%EFTwn-1._dl)&
                            &-CP%EFTwn))*CP%EFTwa)/(a0*(CP%EFTwn-1._dl)*CP%EFTwn))
                    else if (CP%EFTwn==1) then
                        EFTw = a0**(-1._dl -3._dl*(+CP%EFTw0+CP%EFTwa))*exp(3._dl*CP%EFTwa*(a0-1._dl))
                    end if
                end if
            case (4) ! Inflection model
                if (deriv==0) then
                    EFTw = CP%EFTw0 +CP%EFTwa*(CP%EFTwat-a0)**2
                else if (deriv==1) then
                    EFTw = 2._dl*CP%EFTwa*(a0-CP%EFTwat)
                else if (deriv==2) then
                    EFTw = 2._dl*CP%EFTwa
                else if (deriv==3) then
                    EFTw =a0**(-1._dl-3._dl*(CP%EFTw0+CP%EFTwat**2*CP%EFTwa))*&
                        &exp(-1.5_dl*(a0-1._dl)*(1._dl+a0-4._dl*CP%EFTwat)*CP%EFTwa)
                end if
            case (5) ! Taylor expansion
                if (deriv==0) then
                    EFTw = CP%EFTw0 +CP%EFTwa*a0 +0.5_dl*a0**2*CP%EFTw2 + a0**3*CP%EFTw3/6._dl
                else if (deriv==1) then
                    EFTw = CP%EFTwa +CP%EFTw2*a0 +0.5_dl*a0**2*CP%EFTw3
                else if (deriv==2) then
                    EFTw = CP%EFTw2 +CP%EFTw3*a0
                else if (deriv==3) then
                    EFTw = a0**(-1._dl -3._dl*CP%EFTw0)*exp(-(a0-1._dl)/12._dl*(9._dl*(1._dl+a0)&
                        &*CP%EFTw2+2._dl*(1._dl+a0+a0**2)*CP%EFTw3 +36._dl*CP%EFTwa))
                end if
            case (6) ! User defined
                write(*,*) 'EFTCAMB: you have to define EFTw. Go to EFTdeEOS in EFT_main.f90'
                stop
            case default
                write(*,*) 'EFTCAMB: you have to select a model for the dark energy EOS'
                stop
        end select
    end function EFTw

end module EFTdeEOS

! -------------------------------------------------------------------------------------------------

!Solve the designer differential equations for the chosen model and computes,
!from the solution, the values of the EFT functions
module EFTdesigner
    use EFTDef
    use EFTdeEOS
    use MassiveNu
    implicit none
    ! 1) Definitions of the variables common to all the designer code.
    integer , parameter :: des_nstep = 1000        ! Number of points sampled by the designer code.
    integer , parameter :: des_ninterpol = 2       ! Number of points used by the interpolator. Only even.
    real(dl), parameter :: xInitial = log(10._dl**(-8._dl))    ! log(a start)
    real(dl), parameter :: xFinal   = 0._dl                    ! Log(a final)
    integer , parameter :: DesignerInFuture = 0    ! How many points in the future will the code sample
    real(dl), save, dimension(des_nstep+DesignerInFuture+1) :: xp_des   ! Array with the time sampling
    ! 2) Tables containing the sampled values of the EFT functions.
    real(dl), save, dimension(des_nstep+DesignerInFuture+1) :: OmegaDes, OmegapDes, OmegappDes,Omega3pDes
    real(dl), save, dimension(des_nstep+DesignerInFuture+1) :: EFT_Lambda_Des, EFT_Lambdadot_Des, EFT_c_Des, EFT_cdot_Des
    real(dl), save, dimension(des_nstep+DesignerInFuture+1) :: Des_Alpha1, Des_Alpha1p, Des_Alpha2, Des_Alpha2p
    real(dl), save, dimension(des_nstep+DesignerInFuture+1) :: Des_Alpha3, Des_Alpha3p, Des_Alpha4, Des_Alpha4p
    real(dl), save, dimension(des_nstep+DesignerInFuture+1) :: Des_Alpha5, Des_Alpha5p, Des_Alpha6, Des_Alpha6p
    real(dl), save, dimension(des_nstep+DesignerInFuture+1) :: Des_Alpha4pp
    ! 3) Definitions used by a specific model.
    !    1- designer f(R) models:
    integer , parameter  :: nvar_fR = 2
    real(dl), save, dimension(des_nstep+DesignerInFuture+1) :: BFunc, BPrime, fRFunc
    real(dl), save, dimension(des_nstep+DesignerInFuture+1) :: zp_fR ! zp_fR = Ricci/H0^2
    real(dl), save, dimension(nvar_fR, des_nstep+DesignerInFuture+1) :: yp_fR, dyp_fR ! yp_fR = f(R)/H0^2

    logical :: DesignerPrintFlag = .False.

contains

    ! ----------------------------

    subroutine EFTCAMB_Designer(success)
        ! This subroutine calls the correct designer subroutine depending on the selected model.
        implicit none
        logical, intent(inout) :: success

        if (Feedbacklevel>1) write(*,*) 'EFTCAMB: calling designer code'

        select case (CP%DesignerEFTmodel)
            case (1) ! designer f(R) models
                call EFTCAMB_Designer_FR(success)
            case (2) ! designer mc quintessence
                if (CP%EFTwDE==0) write(*,*) 'EFTCAMB: WARNING designer quintessence makes sense only if wDE/=-1'
                success = .true.
            case default
                write(*,*) 'EFTCAMB: wrong choice of designer model'
                stop
        end select
        return
    end subroutine EFTCAMB_Designer

    ! ----------------------------

    subroutine EFTCAMB_Designer_FR(success)
        ! This subroutine finds the appropriate initial conditions and solves
        ! the designer equation for designer f(R) models.
        implicit none
        logical :: success

        real(dl) zbrent

        integer  :: Debug_n=1, Debug_MaxNum
        real(dl) :: B0_Temp, Temp_A, Temp_h, TempMin, TempMax
        integer  :: ind=1 , i=1
        real(dl) :: HorizAsyntB, VertAsyntA, BLeftAsynt, BRightAsynt
        real(dl) :: realAp
        real(dl) :: BTemp1, BTemp2, ATemp1 = 1._dl, Atemp2 = 1._dl, ATemp3, ATemp4, BTemp3, BTemp4

        !  Find initial conditions for designer F(R).
        !  Initial conditions are found solving B0(A)=B0wanted.
        !  The next part of code is complicated by the fact that B0(A) is not continuous and we have to adopt ad-hoc strategies.

        ! 1) Find the horizontal asymptote of B0(A)
        do ind=10, 100,1
            ATemp1 = ATemp2
            ATemp2 = 10._dl**REAL(ind)
            BTemp1 = DesFR_BfuncA(ATemp1)
            BTemp2 = DesFR_BfuncA(ATemp2)
            if (Abs(BTemp1-BTemp2)/Abs(ATemp1-ATemp2).lt.1.d-100) exit
        end do
        HorizAsyntB = BTemp2

        if (FeedbackLevel>2) write(*,*) 'f(R) designer: horizontal asymptote =',  HorizAsyntB

        ! 2) Check that the value of B0 given is not the forbidden one: the one corresponding to the horizontal asynt.
        if (ABS(HorizAsyntB-CP%EFTB0)<1.d-15) then
            success=.false.
            return
        end if

        ! 3) Bracket the vertical asyntote
        ATemp1=-10._dl
        ATemp2=10._dl
        call zbrac(DesFR_BfuncA,ATemp1,ATemp2,success,HorizAsyntB)
        if (.not.success) then
            write(*,*) 'f(R) designer: failure of vert asynt bracketing'
            return
        end if

        ! 4) Find the vertical asyntote by tricking the root finding algorithm.
        VertAsyntA = zbrent(DesFR_BfuncA,ATemp1,ATemp2,1.d-100,HorizAsyntB,success)
        if (.not.success) then
            write(*,*) 'f(R) designer: vertical asyntote not found.'
            return
        end if

        if (FeedbackLevel>2) write(*,*) 'f(R) designer: vertical asymptote =',  VertAsyntA

        ! 5) Find values for A that are on the left and on the right of the asyntote.
        do ind=-10, -1, 1
            ATemp1 = VertAsyntA+10._dl**ind
            ATemp2 = VertAsyntA-10._dl**ind
            BTemp1 = DesFR_BfuncA(ATemp1)
            BTemp2 = DesFR_BfuncA(ATemp2)
            if (BTemp1*BTemp2<0._dl) exit
        end do

        ! 6) Extablish on which side of the asyntote to look for the solution.
        ATemp3 = ATemp1
        ATemp4 = ATemp2
        do ind=1, 10, 1
            ATemp3 = ATemp3 + 10._dl**ind
            ATemp4 = ATemp4 - 10._dl**ind
            BTemp3 = DesFR_BfuncA(ATemp3)
            BTemp4 = DesFR_BfuncA(ATemp4)
            if ((BTemp1-CP%EFTB0)*(BTemp3-CP%EFTB0)<0.or.(BTemp2-CP%EFTB0)*(BTemp4-CP%EFTB0)<0) exit
        end do
        if ((BTemp1-CP%EFTB0)*(BTemp3-CP%EFTB0)<0.and.(BTemp2-CP%EFTB0)*(BTemp4-CP%EFTB0)<0) then
            write(*,*) 'f(R) designer: the root is on both side ???'
            stop
        end if

        ! 7) Solve the equation B0(A)=B0_wanted.
        if ((BTemp1-CP%EFTB0)*(BTemp3-CP%EFTB0)<0) then
            realAp = zbrent(DesFR_BfuncA,ATemp1,ATemp3,1.d-50,CP%EFTB0,success)
            if (.not.success) then
                write(*,*) 'f(R) designer: right side solution not found.'
                stop
            end if
        else if ((BTemp2-CP%EFTB0)*(BTemp4-CP%EFTB0)<0) then
            realAp = zbrent(DesFR_BfuncA,ATemp2,ATemp4,1.d-50,CP%EFTB0,success)
            if (.not.success) then
                write(*,*) 'f(R) designer: right side solution not found'
                stop
            end if
        else
            write(*,*) 'f(R) designer: the root was not on the right side nor the left one...'
            stop
        end if

        if (FeedbackLevel>2) write(*,*) 'f(R) designer: initial condition A =', realAp

        ! 8) Check if the result found is compatible with the requested one. This is required only for debug.
        BTemp1 = DesFR_BfuncA(realAp)
        if (Feedbacklevel > 1) write(*,*) 'f(R) designer: B0 found=', BTemp1, 'B0 given= ',CP%EFTB0

        if (ABS(BTemp1-CP%EFTB0)/ABS(CP%EFTB0)>0.1_dl.and.ABS(BTemp1-CP%EFTB0)>1.d-8.and..false.) then
            write(*,*)  'EFTCAMB: designer code unable to find appropriate initial conditions'
            success = .false.
            return
        end if

        ! 9) Finally: solve the F(R) background equation.
        call DesFR_SolveDesignerEq(realAp)

        ! Code to Debug the f(R) designer code.
        if (DebugDesigner) then
            ! 1) Prints the function B0(A). This is used to debug the initial conditions part.
            call CreateTxtFile('./DebugDesignerB.dat',34)
            print*, 'f(R) designer: Printing B(A) results'
            TempMin     = -10._dl
            TempMax     = +10._dl
            Debug_MaxNum = 1000
            do Debug_n = 1, Debug_MaxNum
                Temp_h = TempMin +REAL(Debug_n-1)*(TempMax-TempMin)/REAL(Debug_MaxNum-1)
                write(34,*) Temp_h, DesFR_BFuncA(Temp_h)
            end do
            close(34)
            ! 2) Prints F(R) quantities.
            DesignerPrintFlag = .True.
            print*, 'f(R) designer: Printing F(R) results'
            call CreateTxtFile('./DebugDesigner.dat',33)
            call DesFR_SolveDesignerEq(realAp)
            close(33)
        end if

        return
    end subroutine EFTCAMB_Designer_FR

    ! ----------------------------

    function DesFR_BfuncA(Ap)
        ! This function gives the present day value of B given the initial conditions.
        implicit none
        real(dl), intent(in) :: Ap
        real(dl) :: DesFR_BfuncA
        call DesFR_SolveDesignerEq(Ap)
        DesFR_BfuncA = BFunc(des_nstep)
        return
    end function DesFR_BfuncA

    ! ----------------------------

    subroutine DesFR_SolveDesignerEq(coefA)
        ! This subroutine solves the designer equation for f(R) models for a given value
        ! of the initial conditions (coefA).
        implicit none
        ! Variable definitions:
        real(dl), intent(in) :: coefA
        real(dl) :: H0_EFT, Omegam_EFT, Omegavac_EFT, OmegaGamma_EFT, OmegaNu_EFT, Omegarad_EFT
        real(dl) :: EquivalenceScale_fR, Ratio_fR, Initial_B_fR, Initial_C_fR, PPlus
        real(dl) :: yStar, CoeffA_Part, yPlus
        real(dl) :: yv_fR(nvar_fR), dyv_fR(nvar_fR)
        real(dl) :: EFT_E_g, EFT_E_gp, EFT_E_gpp, EFT_E_gppp, EFun, EP, EPP, E3P, Ede, Edep, Edepp, fsubRR
        ! Massive neutrinos variables
        real(dl) :: rhonu, presnu, rhonu_tot, presnu_tot, grhormass_t, EFT_E_nu, EFT_EP_nu, EFT_EPP_nu, EFT_E3P_nu
        real(dl) :: presnudot_tot, adotoa, Hdot, presnudotdot_tot, presnudot, presnudotdot
        integer  :: nu_i
        real(dl) :: a, OmegaMassiveNu_EFT

        real(dl) :: x   ! x=log(a) is the time variable.
        real(dl) :: h   ! h is the timestep.
        integer  :: k,i

        real(dl) :: temp1, temp2, temp3, temp4  ! Debug variables

        ! 1) Cosmological parameters:
        H0_EFT = (CP%h0/c_EFT)*1000._dl
        Omegam_EFT = CP%omegab + CP%omegac
        Omegavac_EFT = CP%omegav
        OmegaMassiveNu_EFT = CP%omegan
        OmegaGamma_EFT = kappa_EFT/c_EFT**2*4*sigma_boltz_EFT/c_EFT**3*CP%tcmb**4*Mpc_EFT**2/(3._dl*H0_EFT**2) !8*pi*G/c^2*4*sigma_B/c^3 T^4
        OmegaNu_EFT = 7._dl/8*(4._dl/11)**(4._dl/3)*OmegaGamma_EFT*CP%Num_Nu_massless !7/8*(4/11)^(4/3)*grhog (per neutrino species)
        Omegarad_EFT = OmegaGamma_EFT + OmegaNu_EFT
        EquivalenceScale_fR = (Omegarad_EFT+ OmegaMassiveNu_EFT)/Omegam_EFT
        Ratio_fR = EquivalenceScale_fR/Exp(xInitial)

        ! 2) Growing mode solution:
        Initial_B_fR = (7._dl + 8._dl*Ratio_fR)/(2._dl*(1._dl + Ratio_fR))
        Initial_C_fR = -3._dl/(2._dl*(1._dl + Ratio_fR))
        PPlus = 0.5_dl*(-Initial_B_fR + Sqrt(Initial_B_fR**2 - 4._dl*Initial_C_fR))
        yPlus = exp(PPlus*xInitial)
        !    Construction of the particolar solution:
        CoeffA_Part = (-6._dl*Initial_C_fR)/(-3._dl*Exp(xInitial)*EFTw(Exp(xInitial),1)&
            & +9._dl*EFTw(Exp(xInitial),0)**2&
            & +(18._dl-3._dl*Initial_B_fR)*EFTw(Exp(xInitial),0)&
            & +9._dl -3._dl*Initial_B_fR +Initial_C_fR)
        yStar = CoeffA_Part*Omegavac_EFT*Exp(-2._dl*xInitial)*EFTw(Exp(xInitial),3)

        !    Initial conditions:
        yv_fR(1)=coefA*yPlus + yStar
        yv_fR(2)=PPlus*coefA*yPlus - 3._dl*(1._dl+EFTw(Exp(xInitial),0))*yStar
        xp_des(1)=xInitial
        yp_fR(1,1)=yv_fR(1)
        yp_fR(2,1)=yv_fR(2)
        x=xInitial
        !    Function g(x) at initial time:
        EFT_E_g    = -(Log(EFTw(Exp(x),3)) -2._dl*x)/3._dl
        EFT_E_gp   = 1._dl +EFTw(Exp(x),0)
        EFT_E_gpp  = Exp(x)*EFTw(Exp(x),1)
        EFT_E_gppp = Exp(x)*EFTw(Exp(x),1) +Exp(2._dl*x)*EFTw(Exp(x),2)
        !   Massive neutrinos at initial conditions:
        rhonu_tot  = 0._dl
        presnu_tot = 0._dl
        EFT_E_nu   = 0._dl
        EFT_EP_nu  = 0._dl
        if (CP%Num_Nu_Massive /= 0) then
            do nu_i = 1, CP%Nu_mass_eigenstates
                a= Exp(x)
                rhonu  = 0._dl
                presnu = 0._dl
                grhormass_t=grhormass(nu_i)/a**2
                call Nu_background(a*nu_masses(nu_i),rhonu,presnu)
                rhonu_tot  = rhonu_tot + grhormass_t*rhonu
                presnu_tot = presnu_tot + grhormass_t*presnu

                EFT_E_nu   = EFT_E_nu + grhormass(nu_i)/3._dl/a**4/H0_EFT**2*rhonu
                EFT_EP_nu  = EFT_EP_nu - grhormass(nu_i)/H0_EFT**2/a**4*(rhonu +presnu)
            end do
        end if
        !    Initial values of energy, energy derivative and Ricci.
        EFun = +OmegaRad_EFT*exp(-4._dl*x)&
            & +Omegam_EFT*exp(-3._dl*x)&
            & +Omegavac_EFT*exp(-3._dl*EFT_E_g) + EFT_E_nu
        EP  = -4._dl*OmegaRad_EFT*exp(-4._dl*x)&
            & -3._dl*Omegam_EFT*exp(-3._dl*x)&
            & -3._dl*Omegavac_EFT*EFT_E_gp*exp(-3._dl*EFT_E_g) +EFT_EP_nu
        zp_fR(1) = 3._dl*(4._dl*EFun+EP)

        ! 3) Do the timestep loop:
        h=(xFinal-xInitial)/real(des_nstep-1)
        do k=1, des_nstep+DesignerInFuture

            call  DesFR_derivs(nvar_fR, x, yv_fR, dyv_fR)
            call  EFT_rk4(nvar_fR, yv_fR, dyv_fR, x, h, yv_fR, DesFR_derivs)

            x=x+h
            if (k.le.des_nstep+DesignerInFuture) then
                xp_des(k+1)=x
                yp_fR(1,k+1) = yv_fR(1)
                yp_fR(2,k+1) = yv_fR(2)
                dyp_fR(1,k+1) = dyv_fR(1)
                dyp_fR(2,k+1) = dyv_fR(2)
                ! Compute the g(x) functions
                EFT_E_g    = -(Log(EFTw(Exp(x),3)) -2._dl*x)/3._dl
                EFT_E_gp   = 1._dl +EFTw(Exp(x),0)
                EFT_E_gpp  = Exp(x)*EFTw(Exp(x),1)
                EFT_E_gppp = Exp(x)*EFTw(Exp(x),1) +Exp(2._dl*x)*EFTw(Exp(x),2)
                ! Compute Energy and its derivatives wrt to ln(a):

                ! First compute massive neutrinos contribution:
                rhonu_tot  = 0._dl
                presnu_tot = 0._dl
                EFT_E_nu   = 0._dl
                EFT_EP_nu  = 0._dl
                if (CP%Num_Nu_Massive /= 0) then
                    do nu_i = 1, CP%Nu_mass_eigenstates
                        a= Exp(x)
                        rhonu  = 0._dl
                        presnu = 0._dl
                        grhormass_t=grhormass(nu_i)/a**2
                        call Nu_background(a*nu_masses(nu_i),rhonu,presnu)
                        rhonu_tot  = rhonu_tot + grhormass_t*rhonu
                        presnu_tot = presnu_tot + grhormass_t*presnu

                        EFT_E_nu   = EFT_E_nu + grhormass(nu_i)/3._dl/a**4/H0_EFT**2*rhonu
                        EFT_EP_nu  = EFT_EP_nu - grhormass(nu_i)/H0_EFT**2/a**4*(rhonu +presnu)

                    end do
                end if
                ! Add its contribution to E and E':
                EFun = +OmegaRad_EFT*exp(-4._dl*x)&
                    & +Omegam_EFT*exp(-3._dl*x)&
                    & +Omegavac_EFT*exp(-3._dl*EFT_E_g) + EFT_E_nu
                EP  = -4._dl*OmegaRad_EFT*exp(-4._dl*x)&
                    & -3._dl*Omegam_EFT*exp(-3._dl*x)&
                    & -3._dl*Omegavac_EFT*EFT_E_gp*exp(-3._dl*EFT_E_g) +EFT_EP_nu
                ! Compute everything of massive nu again to get the time derivatives:
                rhonu_tot  = 0._dl
                presnu_tot = 0._dl
                presnudot_tot = 0._dl
                presnudotdot_tot = 0._dl
                EFT_E_nu   = 0._dl
                EFT_EP_nu  = 0._dl
                EFT_EPP_nu = 0._dl
                EFT_E3P_nu = 0._dl
                if (CP%Num_Nu_Massive /= 0) then
                    do nu_i = 1, CP%Nu_mass_eigenstates
                        a= Exp(x)
                        adotoa = +a*H0_EFT*sqrt(EFun)
                        Hdot   = +0.5_dl*H0_EFT**2*a**2*EP +adotoa**2

                        rhonu  = 0._dl
                        presnu = 0._dl
                        presnudot = 0._dl
                        presnudotdot = 0._dl

                        grhormass_t=grhormass(nu_i)/a**2

                        call Nu_background(a*nu_masses(nu_i),rhonu,presnu)
                        presnudot = Nu_pidot(a*nu_masses(nu_i),adotoa,presnu)
                        presnudotdot = Nu_pidotdot(a*nu_masses(nu_i),adotoa,Hdot,presnu,presnudot)

                        rhonu_tot  = rhonu_tot + grhormass_t*rhonu
                        presnu_tot = presnu_tot + grhormass_t*presnu
                        presnudot_tot  = presnudot_tot + grhormass_t*(presnudot -4._dl*adotoa*presnu)
                        presnudotdot_tot = presnudotdot_tot + grhormass_t*(presnudotdot &
                            & -8._dl*adotoa*presnudot +4._dl*presnu*(+4._dl*adotoa**2-Hdot))

                        EFT_E_nu   = EFT_E_nu + grhormass(nu_i)/3._dl/a**4/H0_EFT**2*rhonu
                        EFT_EP_nu  = EFT_EP_nu - grhormass(nu_i)/H0_EFT**2/a**4*(rhonu +presnu)
                        EFT_EPP_nu = EFT_EPP_nu + 3._dl/H0_EFT**2*grhormass(nu_i)/a**4*(rhonu +presnu)&
                            & -grhormass_t*(presnudot -4._dl*adotoa*presnu)/H0_EFT**3/sqrt(EFun)/a**3
                        EFT_E3P_nu = EFT_E3P_nu -9._dl/H0_EFT**2*grhormass(nu_i)/a**4*(rhonu +presnu)&
                            & +(3._dl/adotoa/H0_EFT**2/a**2+Hdot/adotoa**3/H0_EFT**2/a**2)&
                            &*grhormass_t*(presnudot -4._dl*adotoa*presnu)&
                            & -grhormass_t*(presnudotdot &
                            & -8._dl*adotoa*presnudot +4._dl*presnu*(+4._dl*adotoa**2-Hdot))/adotoa**2/H0_EFT**2/a**2
                    end do
                end if

                EPP = 16._dl*OmegaRad_EFT*exp(-4._dl*x)&
                    & +9._dl*Omegam_EFT*exp(-3._dl*x)&
                    & -3._dl*Omegavac_EFT*exp(-3._dl*EFT_E_g)*(EFT_E_gpp-3._dl*EFT_E_gp**2) +EFT_EPP_nu
                E3P = -64._dl*OmegaRad_EFT*exp(-4._dl*x)&
                    & -27._dl*Omegam_EFT*exp(-3._dl*x)&
                    & -3._dl*Omegavac_EFT*exp(-3._dl*EFT_E_g)*(EFT_E_gppp-9._dl*EFT_E_gp*EFT_E_gpp+9._dl*EFT_E_gp**3)&
                    & + EFT_E3P_nu
                Ede= +Omegavac_EFT*exp(-3._dl*EFT_E_g)
                Edep = -3._dl*Omegavac_EFT*EFT_E_gp*exp(-3._dl*EFT_E_g)
                Edepp = -3._dl*Omegavac_EFT*exp(-3._dl*EFT_E_g)*(EFT_E_gpp -3._dl*EFT_E_gp**2)
                ! Compute the Ricci:
                zp_fR(k+1) = 3._dl*(4._dl*EFun+EP)
                ! Compute fsubR:
                fRFunc(k+1) = dyp_fR(1,k+1)/3._dl/(4._dl*EP+EPP)
                ! Compute function B:
                BFunc(k+1) = 2._dl/3._dl/(1._dl+fRFunc(k+1))*1._dl/(4._dl*EP+EPP)*EFun/EP*&
                    &(dyp_fR(2,k+1)-dyp_fR(1,k+1)*(4._dl*EPP+E3P)/(4._dl*EP+EPP))
                ! Compute the EFT functions
                OmegaDes(k+1)   = fRFunc(k+1)
                OmegapDes(k+1)  = Exp(-x)/(6._dl*Efun*(EPP+4._dl*EP))*(-EPP*(6._dl*Ede+yp_fR(1,k+1))&
                    &+EP*(dyp_fR(1,k+1)-4._dl*(6._dl*Ede+yp_fR(1,k+1)))+2._dl*Efun*dyp_fR(1,k+1))
                OmegappDes(k+1) = Exp(-2._dl*x)/(12._dl*Efun**2*(EPP+4._dl*EP))*&
                    &(EP*(EP*(24._dl*Ede-dyp_fR(1,k+1)+4._dl*yp_fR(1,k+1))-6._dl*EFun*(8._dl*Edep+dyp_fR(1,k+1)))&
                    &+EPP*(EP*(6._dl*Ede+yp_fR(1,k+1))-12._dl*Efun*Edep))
                Omega3pDes(k+1) = Exp(-3._dl*x)/(24._dl*Efun**3*(EPP+4._dl*EP))*&
                    &(2._dl*Efun*(EPP**2*(6._dl*Ede+yp_fR(1,k+1))&
                    & +4._dl*EP**2*(18._dl*Edep+6._dl*Ede+2._dl*dyp_fR(1,k+1)+yp_fR(1,k+1))&
                    & +EP*EPP*(18._dl*Edep+30._dl*Ede-dyp_fR(1,k+1)+5._dl*yp_fR(1,k+1)))&
                    & +12._dl*Efun**2*(EPP*(-2._dl*Edepp+4._dl*Edep-dyp_fR(1,k+1))&
                    & +EP*(-8._dl*Edepp+16._dl*Edep+dyp_fR(1,k+1)))&
                    & +3._dl*EP**2*(EP*(dyp_fR(1,k+1)-4._dl*(6._dl*Ede+yp_fR(1,k+1)))-EPP*(6._dl*Ede+yp_fR(1,k+1))))
                EFT_Lambda_Des(k+1)    = 0.5_dl*H0_EFT**2*(yp_fR(1,k+1)-fRFunc(k+1)*zp_fR(k+1))*Exp(x)**2
                EFT_Lambdadot_Des(k+1) = -1.5_dl*H0_EFT**3*Sqrt(Efun)*(Exp(x)**4*OmegapDes(k+1)*(4._dl*Efun+EP))
                ! Compute other f(R) quantities:
                fsubRR = 0.5_dl*EP/Efun/3._dl*BFunc(k+1)*(1._dl+fRFunc(k+1))/(4._dl*EP+EPP)
                ! Debug output:
                if (DebugDesigner.and.DesignerPrintFlag) then
                !     write (33,'(20E15.5)') x, Exp(x), zp_fR(k+1), yp_fR(1,k+1), &
                !         &OmegaDes(k+1), OmegapDes(k+1), OmegappDes(k+1), Omega3pDes(k+1), EFT_Lambda_Des(k+1), EFT_Lambdadot_Des(k+1),&
                !         &BFunc(k+1), fsubRR
                end if
            end if
        end do

        return
    end subroutine DesFR_SolveDesignerEq

    ! ----------------------------

    subroutine DesFR_derivs(n,x,y_fR,dydx_fR)
        ! This subroutine computes the designer f(R) equation of motion.
        implicit none
        integer n
        real(dl) :: x
        real(dl), dimension(n) :: y_fR,dydx_fR
        real(dl) :: EFunction, EFunPrime, EFunPrime2, EFunPrime3
        real(dl) :: EFTwde0, EFTwde1, EFTwde2, EFTwde3
        real(dl) :: EFT_E_gfun, EFT_E_gfunp, EFT_E_gfunpp, EFT_E_gfunppp
        real(dl) :: H0_EFT, Omegam_EFT, Omegavac_EFT, OmegaGamma_EFT, OmegaNu_EFT, Omegarad_EFT
        ! Massive neutrinos variables
        real(dl) :: rhonu, presnu, rhonu_tot, presnu_tot, grhormass_t, EFT_E_nu, EFT_EP_nu, EFT_EPP_nu, EFT_E3P_nu
        real(dl) :: presnudot_tot, adotoa, Hdot, presnudotdot_tot, presnudot, presnudotdot
        integer  :: nu_i
        real(dl) :: a
        ! 1) Cosmological parameters:
        H0_EFT = (CP%h0/c_EFT)*1000._dl
        Omegam_EFT = CP%omegab + CP%omegac
        Omegavac_EFT = CP%omegav
        OmegaGamma_EFT = kappa_EFT/c_EFT**2*4*sigma_boltz_EFT/c_EFT**3*CP%tcmb**4*Mpc_EFT**2/(3._dl*H0_EFT**2) !8*pi*G/c^2*4*sigma_B/c^3 T^4
        OmegaNu_EFT = 7._dl/8*(4._dl/11)**(4._dl/3)*OmegaGamma_EFT !7/8*(4/11)^(4/3)*grhog (per neutrino species)
        Omegarad_EFT = OmegaGamma_EFT + OmegaNu_EFT
        ! 2) Compute the function g(x) and its derivatives:
        EFT_E_gfun    = -(Log(EFTw(Exp(x),3)) -2._dl*x)/3._dl
        EFT_E_gfunp   = 1._dl +EFTw(Exp(x),0)
        EFT_E_gfunpp  = Exp(x)*EFTw(Exp(x),1)
        EFT_E_gfunppp = Exp(x)*EFTw(Exp(x),1) +Exp(2._dl*x)*EFTw(Exp(x),2)
        ! 3) Compute energy and its derivatives:

        ! First compute massive neutrinos contribution:
        rhonu_tot  = 0._dl
        presnu_tot = 0._dl
        EFT_E_nu   = 0._dl
        EFT_EP_nu  = 0._dl
        if (CP%Num_Nu_Massive /= 0) then
            do nu_i = 1, CP%Nu_mass_eigenstates
                a= Exp(x)
                rhonu  = 0._dl
                presnu = 0._dl
                grhormass_t=grhormass(nu_i)/a**2
                call Nu_background(a*nu_masses(nu_i),rhonu,presnu)
                rhonu_tot  = rhonu_tot + grhormass_t*rhonu
                presnu_tot = presnu_tot + grhormass_t*presnu

                EFT_E_nu   = EFT_E_nu + grhormass(nu_i)/3._dl/a**4/H0_EFT**2*rhonu
                EFT_EP_nu  = EFT_EP_nu - grhormass(nu_i)/H0_EFT**2/a**4*(rhonu +presnu)

            end do
        end if

        ! Add its contribution to E and E':

        EFunction = +OmegaRad_EFT*exp(-4._dl*x)&
            & +Omegam_EFT*exp(-3._dl*x)&
            & +Omegavac_EFT*exp(-3._dl*EFT_E_gfun) + EFT_E_nu
        EFunPrime = -4._dl*OmegaRad_EFT*exp(-4._dl*x)&
            & -3._dl*Omegam_EFT*exp(-3._dl*x)&
            & -3._dl*Omegavac_EFT*EFT_E_gfunp*exp(-3._dl*EFT_E_gfun) +EFT_EP_nu

        ! Compute everything of massive nu again to get the time derivatives:
        rhonu_tot  = 0._dl
        presnu_tot = 0._dl
        presnudot_tot = 0._dl
        presnudotdot_tot = 0._dl
        EFT_E_nu   = 0._dl
        EFT_EP_nu  = 0._dl
        EFT_EPP_nu = 0._dl
        EFT_E3P_nu = 0._dl
        if (CP%Num_Nu_Massive /= 0) then
            do nu_i = 1, CP%Nu_mass_eigenstates
                a= Exp(x)
                adotoa = +a*H0_EFT*sqrt(EFunction)
                Hdot   = +0.5_dl*H0_EFT**2*a**2*EFunPrime +adotoa**2

                rhonu  = 0._dl
                presnu = 0._dl
                presnudot = 0._dl
                presnudotdot = 0._dl

                grhormass_t=grhormass(nu_i)/a**2

                call Nu_background(a*nu_masses(nu_i),rhonu,presnu)
                presnudot = Nu_pidot(a*nu_masses(nu_i),adotoa,presnu)
                presnudotdot = Nu_pidotdot(a*nu_masses(nu_i),adotoa,Hdot,presnu,presnudot)

                rhonu_tot  = rhonu_tot + grhormass_t*rhonu
                presnu_tot = presnu_tot + grhormass_t*presnu
                presnudot_tot  = presnudot_tot + grhormass_t*(presnudot -4._dl*adotoa*presnu)
                presnudotdot_tot = presnudotdot_tot + grhormass_t*(presnudotdot &
                    & -8._dl*adotoa*presnudot +4._dl*presnu*(+4._dl*adotoa**2-Hdot))

                EFT_E_nu   = EFT_E_nu + grhormass(nu_i)/3._dl/a**4/H0_EFT**2*rhonu
                EFT_EP_nu  = EFT_EP_nu - grhormass(nu_i)/H0_EFT**2/a**4*(rhonu +presnu)
                EFT_EPP_nu = EFT_EPP_nu + 3._dl/H0_EFT**2*grhormass(nu_i)/a**4*(rhonu +presnu)&
                    & -grhormass_t*(presnudot -4._dl*adotoa*presnu)/H0_EFT**3/sqrt(EFunction)/a**3
                EFT_E3P_nu = EFT_E3P_nu -9._dl/H0_EFT**2*grhormass(nu_i)/a**4*(rhonu +presnu)&
                    & +(3._dl/adotoa/H0_EFT**2/a**2+Hdot/adotoa**3/H0_EFT**2/a**2)&
                    &*grhormass_t*(presnudot -4._dl*adotoa*presnu)&
                    & -grhormass_t*(presnudotdot &
                    & -8._dl*adotoa*presnudot +4._dl*presnu*(+4._dl*adotoa**2-Hdot))/adotoa**2/H0_EFT**2/a**2
            end do
        end if

        EFunPrime2 = 16._dl*OmegaRad_EFT*exp(-4._dl*x)&
            & +9._dl*Omegam_EFT*exp(-3._dl*x)&
            & -3._dl*Omegavac_EFT*exp(-3._dl*EFT_E_gfun)*(EFT_E_gfunpp -3._dl*EFT_E_gfunp**2) + EFT_EPP_nu
        EFunPrime3 = -64._dl*OmegaRad_EFT*exp(-4._dl*x)&
            & -27._dl*Omegam_EFT*exp(-3._dl*x)&
            & -3._dl*Omegavac_EFT*exp(-3._dl*EFT_E_gfun)*&
            &(EFT_E_gfunppp-9._dl*EFT_E_gfunp*EFT_E_gfunpp+9._dl*EFT_E_gfunp**3) + EFT_E3P_nu

        ! 4) Get the equation of motion:
        dydx_fR(1) = y_fR(2)
        dydx_fR(2) = (1._dl+0.5_dl*EFunPrime/EFunction+(4._dl*EFunPrime2+EFunPrime3)/(4._dl*EFunPrime+EFunPrime2))*y_fR(2) &
            & -0.5_dl*(4._dl*EFunPrime+EFunPrime2)/EFunction*y_fR(1) &
            & -3._dl*Omegavac_EFT*exp(-3._dl*EFT_E_gfun)*(4._dl*EFunPrime+EFunPrime2)/EFunction
        return
    end subroutine DesFR_derivs

    ! ----------------------------

    function Designed_EFT_Function(EFTFunctionTable, a, GR_value)
        ! The designer code will provide a table of sampled values for the EFT functions.
        ! This function is called in the EFTfunctions module to interpolate those tables.
        implicit none
        real(dl), intent(in) :: a, GR_value
        real(dl) :: EFTFunctionTable(des_nstep+DesignerInFuture+1)
        real(dl) :: Designed_EFT_Function

        real(dl) :: x, temp, dtemp
        real(dl) :: xb(des_ninterpol),yb(des_ninterpol)
        integer  :: i, jlo, stint

        x = log(a)

        if (x.lt.xp_des(1)) then
            temp = GR_value

        else if (x.ge.xp_des(1) .and. x.le.xp_des(des_nstep+DesignerInFuture)) then
            ! 1) find the point corresponing to the requested input.
            call hunt(xp_des, des_nstep+DesignerInFuture, x, jlo)
            ! 2) construct the table that will be interpolated
            stint = des_ninterpol/2
            if((des_nstep+DesignerInFuture-jlo-stint).lt.0) then
                ! Last des_ninterpol points
                do i=1, des_ninterpol
                    xb(i)=xp_des(des_nstep+DesignerInFuture-des_ninterpol+i)
                    yb(i)=EFTFunctionTable(des_nstep+DesignerInFuture-des_ninterpol+i)
                end do
            else if (jlo.eq.1) then
                ! First des_ninterpol points
                do i=1, des_ninterpol
                    xb(i)=xp_des(i)
                    yb(i)=EFTFunctionTable(i)
                end do
            else
                ! Surrounding des_ninterpol points
                do i=1, des_ninterpol
                    xb(i)=xp_des(jlo-stint+i)
                    yb(i)=EFTFunctionTable(jlo-stint+i)
                end do
            endif
            ! 3) Call the Neville interpolator.
            call Polint(des_ninterpol, xb, yb, x, temp, dtemp)

        else
            temp = EFTFunctionTable(des_nstep+DesignerInFuture)

        end if

        Designed_EFT_Function = temp

    end function Designed_EFT_Function

end module EFTdesigner

! -------------------------------------------------------------------------------------------------

!module EFTmapping
!   Code used to construct the mapping of a specific model into the EFT framework.
!end module EFTmapping

! -------------------------------------------------------------------------------------------------

!Definitions of the EFT functions later used by the code.
!After the designer or mapping code have been called the EFT functions can be used directly from
!this module.
module EFTfunctions
    use EFTDef
    use EFTdeEOS
    use EFTdesigner
contains

    ! ----------------------------

    function EFTOmega(a,deriv)
        !EFT function Omega
        implicit none
        real(dl) :: EFTOmega
        real(dl) , intent(in) :: a      ! scale factor
        integer  , intent(in) :: deriv  ! order of derivative wrt the scale factor

        select case (CP%EFTflag)
            case (0) ! GR
                EFTOmega = 0._dl
            case (1) ! Pure EFT
                select case (CP%PureEFTmodelOmega)
                    case (0) ! 0) Operator neglected;
                        EFTOmega = 0._dl
                    case (1) ! 1) Constant model;
                        if (deriv==0) then
                            EFTOmega = CP%EFTOmega0
                        else
                            EFTOmega = 0._dl
                        end if
                    case (2) ! 2) Linear model;
                        if (deriv==0) then
                            EFTOmega= CP%EFTOmega0*a
                        else if (deriv==1) then
                            EFTOmega= CP%EFTOmega0
                        else
                            EFTOmega= 0._dl
                        end if
                    case (3) ! 3) Power law model;
                        if (deriv==0) then
                            EFTOmega= CP%EFTOmega0*a**CP%EFTOmegaExp
                        else if (deriv==1) then
                            EFTOmega= CP%EFTOmegaExp*CP%EFTOmega0*a**(CP%EFTOmegaExp-1.0_dl)
                        else if (deriv==2) then
                            EFTOmega= (CP%EFTOmegaExp-1.0_dl)*CP%EFTOmegaExp*CP%EFTOmega0*&
                                &a**(CP%EFTOmegaExp-2.0_dl)
                        else if (deriv==3) then
                            EFTOmega= (CP%EFTOmegaExp-2.0_dl)*(CP%EFTOmegaExp-1.0_dl)*&
                                &CP%EFTOmegaExp*CP%EFTOmega0*a**(CP%EFTOmegaExp-3.0_dl)
                        end if
                    case (4) ! 4) Exponential model;
                        if (deriv==0) then
                            EFTOmega= Exp(CP%EFTOmega0*a**CP%EFTOmegaExp)-1._dl
                        else if (deriv==1) then
                            EFTOmega= CP%EFTOmegaExp*CP%EFTOmega0*a**(CP%EFTOmegaExp-1.0_dl)&
                                &*Exp(CP%EFTOmega0*a**CP%EFTOmegaExp)
                        else if (deriv==2) then
                            EFTOmega= a**(-2._dl+CP%EFTOmegaExp)*Exp(a**CP%EFTOmegaExp*CP%EFTOmega0)*&
                                &CP%EFTOmega0*CP%EFTOmegaExp*(-1._dl+CP%EFTOmegaExp&
                                &+a**CP%EFTOmegaExp*CP%EFTOmega0*CP%EFTOmegaExp)
                        else if (deriv==3) then
                            EFTOmega= a**(-3._dl+CP%EFTOmegaExp)*Exp(a**CP%EFTOmegaExp*CP%EFTOmega0)*&
                                &CP%EFTOmega0*CP%EFTOmegaExp*(2._dl-3._dl*(1._dl+a**CP%EFTOmegaExp*&
                                &CP%EFTOmega0)*CP%EFTOmegaExp+(1._dl+3._dl*a**CP%EFTOmegaExp*&
                                &CP%EFTOmega0+a**(2._dl*CP%EFTOmegaExp)*CP%EFTOmega0**2)&
                                &*CP%EFTOmegaExp**2)
                        end if
                    case (5) ! 5) User defined;
                        stop 'You have to define the model for EFTOmega first!'
                        if (deriv==0) then
                        else if (deriv==1) then
                        else if (deriv==2) then
                        else if (deriv==3) then
                        end if
                    case default
                        stop 'Wrong model for EFTOmega'
                end select
            case (2) ! Designer mapping EFT
                select case (CP%DesignerEFTModel)
                    case (1) ! Designer f(R) models
                        if (deriv==0) then
                            EFTOmega= Designed_EFT_Function(OmegaDes, a, 0._dl)
                        else if (deriv==1) then
                            EFTOmega= Designed_EFT_Function(OmegapDes, a, 0._dl)
                        else if (deriv==2) then
                            EFTOmega= Designed_EFT_Function(OmegappDes, a, 0._dl)
                        else if (deriv==3) then
                            EFTOmega= Designed_EFT_Function(Omega3pDes, a, 0._dl)
                        end if
                    case (2) ! Designer mc quintessence
                        EFTOmega = 0._dl
                    case default
                        stop 'Wrong model for EFTOmega'
                end select
            case default
                stop 'Wrong model for EFTOmega'
        end select

    end function EFTOmega

    ! ----------------------------

    function EFTAlpha1(a,deriv)
        ! EFT function Alpha_1
        implicit none
        real(dl) :: EFTAlpha1
        real(dl) , intent(in) :: a      ! scale factor
        integer  , intent(in) :: deriv  ! order of derivative wrt the scale factor

        select case (CP%EFTflag)
            case (0) ! GR
                EFTAlpha1 = 0._dl
            case (1) ! Pure EFT
                select case (CP%PureEFTmodelAlpha1)
                    case (0) ! 0) Operator neglected;
                        EFTAlpha1 = 0._dl
                    case (1) ! 1) Constant model;
                        if (deriv==0) then
                            EFTAlpha1 = CP%EFTAlpha10
                        else
                            EFTAlpha1 = 0._dl
                        end if
                    case (2) ! 2) Linear model;
                        if (deriv==0) then
                            EFTAlpha1= CP%EFTAlpha10*a
                        else if (deriv==1) then
                            EFTAlpha1= CP%EFTAlpha10
                        else
                            EFTAlpha1= 0._dl
                        end if
                    case (3) ! 3) Power law model;
                        if (deriv==0) then
                            EFTAlpha1= CP%EFTAlpha10*a**CP%EFTAlpha1Exp
                        else if (deriv==1) then
                            EFTAlpha1= CP%EFTAlpha1Exp*CP%EFTAlpha10*a**(CP%EFTAlpha1Exp-1.0_dl)
                        else if (deriv==2) then
                            EFTAlpha1= (CP%EFTAlpha1Exp-1.0_dl)*CP%EFTAlpha1Exp*CP%EFTAlpha10*&
                                &a**(CP%EFTAlpha1Exp-2.0_dl)
                        else if (deriv==3) then
                            EFTAlpha1= (CP%EFTAlpha1Exp-2.0_dl)*(CP%EFTAlpha1Exp-1.0_dl)*&
                                &CP%EFTAlpha1Exp*CP%EFTAlpha10*a**(CP%EFTAlpha1Exp-3.0_dl)
                        end if
                    case (4) ! 4) Exponential model;
                        if (deriv==0) then
                            EFTAlpha1= Exp(CP%EFTAlpha10*a**CP%EFTAlpha1Exp)-1._dl
                        else if (deriv==1) then
                            EFTAlpha1= CP%EFTAlpha1Exp*CP%EFTAlpha10*a**(CP%EFTAlpha1Exp-1.0_dl)&
                                &*Exp(CP%EFTAlpha10*a**CP%EFTAlpha1Exp)
                        else if (deriv==2) then
                            EFTAlpha1= a**(-2._dl+CP%EFTAlpha1Exp)*Exp(a**CP%EFTAlpha1Exp*CP%EFTAlpha10)*&
                                &CP%EFTAlpha10*CP%EFTAlpha1Exp*(-1._dl+CP%EFTAlpha1Exp&
                                &+a**CP%EFTAlpha1Exp*CP%EFTAlpha10*CP%EFTAlpha1Exp)
                        else if (deriv==3) then
                            EFTAlpha1= a**(-3._dl+CP%EFTAlpha1Exp)*Exp(a**CP%EFTAlpha1Exp*CP%EFTAlpha10)*&
                                &CP%EFTAlpha10*CP%EFTAlpha1Exp*(2._dl-3._dl*(1._dl+a**CP%EFTAlpha1Exp*&
                                &CP%EFTAlpha10)*CP%EFTAlpha1Exp+(1._dl+3._dl*a**CP%EFTAlpha1Exp*&
                                &CP%EFTAlpha10+a**(2._dl*CP%EFTAlpha1Exp)*CP%EFTAlpha10**2)&
                                &*CP%EFTAlpha1Exp**2)
                        end if
                    case (5) ! 5) User defined;
                        stop 'You have to define the model for EFTAlpha1 first!'
                        if (deriv==0) then
                        else if (deriv==1) then
                        else if (deriv==2) then
                        else if (deriv==3) then
                        end if
                    case default
                        stop 'Wrong choice for EFTAlpha1'
                end select
            case (2) ! Designer mapping EFT
                select case (CP%DesignerEFTModel)
                    case (1) ! Designer f(R) models
                        EFTAlpha1=0._dl
                    case (2) ! Designer mc quintessence
                        EFTAlpha1=0._dl
                    case default
                        stop 'Wrong choice for EFTAlpha1'
                end select
            case default
                stop 'Wrong choice for EFTAlpha1'
        end select

    end function EFTAlpha1

    ! ----------------------------

    function EFTAlpha2(a,deriv)
        ! EFT function Alpha_2
        implicit none
        real(dl) :: EFTAlpha2
        real(dl) , intent(in) :: a      ! scale factor
        integer  , intent(in) :: deriv  ! order of derivative wrt the scale factor

        select case (CP%EFTflag)
            case (0) ! GR
                EFTAlpha2 = 0._dl
            case (1) ! Pure EFT
                select case (CP%PureEFTmodelAlpha2)
                    case (0) ! 0) Operator neglected;
                        EFTAlpha2 = 0._dl
                    case (1) ! 1) Constant model;
                        if (deriv==0) then
                            EFTAlpha2 = CP%EFTAlpha20
                        else
                            EFTAlpha2 = 0._dl
                        end if
                    case (2) ! 2) Linear model;
                        if (deriv==0) then
                            EFTAlpha2= CP%EFTAlpha20*a
                        else if (deriv==1) then
                            EFTAlpha2= CP%EFTAlpha20
                        else
                            EFTAlpha2= 0._dl
                        end if
                    case (3) ! 3) Power law model;
                        if (deriv==0) then
                            EFTAlpha2= CP%EFTAlpha20*a**CP%EFTAlpha2Exp
                        else if (deriv==1) then
                            EFTAlpha2= CP%EFTAlpha2Exp*CP%EFTAlpha20*a**(CP%EFTAlpha2Exp-1.0_dl)
                        else if (deriv==2) then
                            EFTAlpha2= (CP%EFTAlpha2Exp-1.0_dl)*CP%EFTAlpha2Exp*CP%EFTAlpha20*&
                                &a**(CP%EFTAlpha2Exp-2.0_dl)
                        else if (deriv==3) then
                            EFTAlpha2= (CP%EFTAlpha2Exp-2.0_dl)*(CP%EFTAlpha2Exp-1.0_dl)*&
                                &CP%EFTAlpha2Exp*CP%EFTAlpha20*a**(CP%EFTAlpha2Exp-3.0_dl)
                        end if
                    case (4) ! 4) Exponential model;
                        if (deriv==0) then
                            EFTAlpha2= Exp(CP%EFTAlpha20*a**CP%EFTAlpha2Exp)-1._dl
                        else if (deriv==1) then
                            EFTAlpha2= CP%EFTAlpha2Exp*CP%EFTAlpha20*a**(CP%EFTAlpha2Exp-1.0_dl)&
                                &*Exp(CP%EFTAlpha20*a**CP%EFTAlpha2Exp)
                        else if (deriv==2) then
                            EFTAlpha2= a**(-2._dl+CP%EFTAlpha2Exp)*Exp(a**CP%EFTAlpha2Exp*CP%EFTAlpha20)*&
                                &CP%EFTAlpha20*CP%EFTAlpha2Exp*(-1._dl+CP%EFTAlpha2Exp&
                                &+a**CP%EFTAlpha2Exp*CP%EFTAlpha20*CP%EFTAlpha2Exp)
                        else if (deriv==3) then
                            EFTAlpha2= a**(-3._dl+CP%EFTAlpha2Exp)*Exp(a**CP%EFTAlpha2Exp*CP%EFTAlpha20)*&
                                &CP%EFTAlpha20*CP%EFTAlpha2Exp*(2._dl-3._dl*(1._dl+a**CP%EFTAlpha2Exp*&
                                &CP%EFTAlpha20)*CP%EFTAlpha2Exp+(1._dl+3._dl*a**CP%EFTAlpha2Exp*&
                                &CP%EFTAlpha20+a**(2._dl*CP%EFTAlpha2Exp)*CP%EFTAlpha20**2)&
                                &*CP%EFTAlpha2Exp**2)
                        end if
                    case (5) ! 5) User defined;
                        stop 'You have to define the model for EFTAlpha2 first!'
                        if (deriv==0) then
                        else if (deriv==1) then
                        else if (deriv==2) then
                        else if (deriv==3) then
                        end if
                    case default
                        stop 'Wrong choice for EFTAlpha2'
                end select
            case (2) ! Designer mapping EFT
                select case (CP%DesignerEFTModel)
                    case (1) ! Designer f(R) models
                        EFTAlpha2=0._dl
                    case (2) ! Designer mc quintessence
                        EFTAlpha2=0._dl
                    case default
                        stop 'Wrong choice for EFTAlpha2'
                end select
            case default
                stop 'Wrong choice for EFTAlpha2'
        end select

    end function EFTAlpha2

    ! ----------------------------

    function EFTAlpha3(a,deriv)
        ! EFT function Alpha_3
        implicit none
        real(dl) :: EFTAlpha3
        real(dl) , intent(in) :: a      ! scale factor
        integer  , intent(in) :: deriv  ! order of derivative wrt the scale factor

        select case (CP%EFTflag)
            case (0) ! GR
                EFTAlpha3 = 0._dl
            case (1) ! Pure EFT
                select case (CP%PureEFTmodelAlpha3)
                    case (0) ! 0) Operator neglected;
                        EFTAlpha3 = 0._dl
                    case (1) ! 1) Constant model;
                        if (deriv==0) then
                            EFTAlpha3 = CP%EFTAlpha30
                        else
                            EFTAlpha3 = 0._dl
                        end if
                    case (2) ! 2) Linear model;
                        if (deriv==0) then
                            EFTAlpha3= CP%EFTAlpha30*a
                        else if (deriv==1) then
                            EFTAlpha3= CP%EFTAlpha30
                        else
                            EFTAlpha3= 0._dl
                        end if
                    case (3) ! 3) Power law model;
                        if (deriv==0) then
                            EFTAlpha3= CP%EFTAlpha30*a**CP%EFTAlpha3Exp
                        else if (deriv==1) then
                            EFTAlpha3= CP%EFTAlpha3Exp*CP%EFTAlpha30*a**(CP%EFTAlpha3Exp-1.0_dl)
                        else if (deriv==2) then
                            EFTAlpha3= (CP%EFTAlpha3Exp-1.0_dl)*CP%EFTAlpha3Exp*CP%EFTAlpha30*&
                                &a**(CP%EFTAlpha3Exp-2.0_dl)
                        else if (deriv==3) then
                            EFTAlpha3= (CP%EFTAlpha3Exp-2.0_dl)*(CP%EFTAlpha3Exp-1.0_dl)*&
                                &CP%EFTAlpha3Exp*CP%EFTAlpha30*a**(CP%EFTAlpha3Exp-3.0_dl)
                        end if
                    case (4) ! 4) Exponential model;
                        if (deriv==0) then
                            EFTAlpha3= Exp(CP%EFTAlpha30*a**CP%EFTAlpha3Exp)-1._dl
                        else if (deriv==1) then
                            EFTAlpha3= CP%EFTAlpha3Exp*CP%EFTAlpha30*a**(CP%EFTAlpha3Exp-1.0_dl)&
                                &*Exp(CP%EFTAlpha30*a**CP%EFTAlpha3Exp)
                        else if (deriv==2) then
                            EFTAlpha3= a**(-2._dl+CP%EFTAlpha3Exp)*Exp(a**CP%EFTAlpha3Exp*CP%EFTAlpha30)*&
                                &CP%EFTAlpha30*CP%EFTAlpha3Exp*(-1._dl+CP%EFTAlpha3Exp&
                                &+a**CP%EFTAlpha3Exp*CP%EFTAlpha30*CP%EFTAlpha3Exp)
                        else if (deriv==3) then
                            EFTAlpha3= a**(-3._dl+CP%EFTAlpha3Exp)*Exp(a**CP%EFTAlpha3Exp*CP%EFTAlpha30)*&
                                &CP%EFTAlpha30*CP%EFTAlpha3Exp*(2._dl-3._dl*(1._dl+a**CP%EFTAlpha3Exp*&
                                &CP%EFTAlpha30)*CP%EFTAlpha3Exp+(1._dl+3._dl*a**CP%EFTAlpha3Exp*&
                                &CP%EFTAlpha30+a**(2._dl*CP%EFTAlpha3Exp)*CP%EFTAlpha30**2)&
                                &*CP%EFTAlpha3Exp**2)
                        end if
                    case (5) ! 5) User defined;
                        stop 'You have to define the model for EFTAlpha3 first!'
                        if (deriv==0) then
                        else if (deriv==1) then
                        else if (deriv==2) then
                        else if (deriv==3) then
                        end if
                    case default
                        stop 'Wrong choice for EFTAlpha3'
                end select
            case (2) ! Designer mapping EFT
                select case (CP%DesignerEFTModel)
                    case (1) ! Designer f(R) models
                        EFTAlpha3=0._dl
                    case (2) ! Designer mc quintessence
                        EFTAlpha3=0._dl
                    case default
                        stop 'Wrong choice for EFTAlpha3'
                end select
            case default
                stop 'Wrong choice for EFTAlpha3'
        end select

    end function EFTAlpha3

    ! ----------------------------

    function EFTAlpha4(a,deriv)
        ! EFT function Alpha_4
        implicit none
        real(dl) :: EFTAlpha4
        real(dl) , intent(in) :: a      ! scale factor
        integer  , intent(in) :: deriv  ! order of derivative wrt the scale factor

        select case (CP%EFTflag)
            case (0) ! GR
                EFTAlpha4 = 0._dl
            case (1) ! Pure EFT
                select case (CP%PureEFTmodelAlpha4)
                    case (0) ! 0) Operator neglected;
                        EFTAlpha4 = 0._dl
                    case (1) ! 1) Constant model;
                        if (deriv==0) then
                            EFTAlpha4 = CP%EFTAlpha40
                        else
                            EFTAlpha4 = 0._dl
                        end if
                    case (2) ! 2) Linear model;
                        if (deriv==0) then
                            EFTAlpha4= CP%EFTAlpha40*a
                        else if (deriv==1) then
                            EFTAlpha4= CP%EFTAlpha40
                        else
                            EFTAlpha4= 0._dl
                        end if
                    case (3) ! 3) Power law model;
                        if (deriv==0) then
                            EFTAlpha4= CP%EFTAlpha40*a**CP%EFTAlpha4Exp
                        else if (deriv==1) then
                            EFTAlpha4= CP%EFTAlpha4Exp*CP%EFTAlpha40*a**(CP%EFTAlpha4Exp-1.0_dl)
                        else if (deriv==2) then
                            EFTAlpha4= (CP%EFTAlpha4Exp-1.0_dl)*CP%EFTAlpha4Exp*CP%EFTAlpha40*&
                                &a**(CP%EFTAlpha4Exp-2.0_dl)
                        else if (deriv==3) then
                            EFTAlpha4= (CP%EFTAlpha4Exp-2.0_dl)*(CP%EFTAlpha4Exp-1.0_dl)*&
                                &CP%EFTAlpha4Exp*CP%EFTAlpha40*a**(CP%EFTAlpha4Exp-3.0_dl)
                        end if
                    case (4) ! 4) Exponential model;
                        if (deriv==0) then
                            EFTAlpha4= Exp(CP%EFTAlpha40*a**CP%EFTAlpha4Exp)-1._dl
                        else if (deriv==1) then
                            EFTAlpha4= CP%EFTAlpha4Exp*CP%EFTAlpha40*a**(CP%EFTAlpha4Exp-1.0_dl)&
                                &*Exp(CP%EFTAlpha40*a**CP%EFTAlpha4Exp)
                        else if (deriv==2) then
                            EFTAlpha4= a**(-2._dl+CP%EFTAlpha4Exp)*Exp(a**CP%EFTAlpha4Exp*CP%EFTAlpha40)*&
                                &CP%EFTAlpha40*CP%EFTAlpha4Exp*(-1._dl+CP%EFTAlpha4Exp&
                                &+a**CP%EFTAlpha4Exp*CP%EFTAlpha40*CP%EFTAlpha4Exp)
                        else if (deriv==3) then
                            EFTAlpha4= a**(-3._dl+CP%EFTAlpha4Exp)*Exp(a**CP%EFTAlpha4Exp*CP%EFTAlpha40)*&
                                &CP%EFTAlpha40*CP%EFTAlpha4Exp*(2._dl-3._dl*(1._dl+a**CP%EFTAlpha4Exp*&
                                &CP%EFTAlpha40)*CP%EFTAlpha4Exp+(1._dl+3._dl*a**CP%EFTAlpha4Exp*&
                                &CP%EFTAlpha40+a**(2._dl*CP%EFTAlpha4Exp)*CP%EFTAlpha40**2)&
                                &*CP%EFTAlpha4Exp**2)
                        end if
                    case (5) ! 5) User defined;
                        stop 'You have to define the model for EFTAlpha4 first!'
                        if (deriv==0) then
                        else if (deriv==1) then
                        else if (deriv==2) then
                        else if (deriv==3) then
                        end if
                    case default
                        stop 'Wrong choice for EFTAlpha4'
                end select
            case (2) ! Designer mapping EFT
                select case (CP%DesignerEFTModel)
                    case (1) ! Designer f(R) models
                        EFTAlpha4=0._dl
                    case (2) ! Designer mc quintessence
                        EFTAlpha4=0._dl
                    case default
                        stop 'Wrong choice for EFTAlpha4'
                end select
            case default
                stop 'Wrong choice for EFTAlpha4'
        end select

    end function EFTAlpha4

    ! ----------------------------

    function EFTAlpha5(a,deriv)
        ! EFT function Alpha_5
        implicit none
        real(dl) :: EFTAlpha5
        real(dl) , intent(in) :: a      ! scale factor
        integer  , intent(in) :: deriv  ! order of derivative wrt the scale factor

        select case (CP%EFTflag)
            case (0) ! GR
                EFTAlpha5 = 0._dl
            case (1) ! Pure EFT
                select case (CP%PureEFTmodelAlpha5)
                    case (0) ! 0) Operator neglected;
                        EFTAlpha5 = 0._dl
                    case (1) ! 1) Constant model;
                        if (deriv==0) then
                            EFTAlpha5 = CP%EFTAlpha50
                        else
                            EFTAlpha5 = 0._dl
                        end if
                    case (2) ! 2) Linear model;
                        if (deriv==0) then
                            EFTAlpha5= CP%EFTAlpha50*a
                        else if (deriv==1) then
                            EFTAlpha5= CP%EFTAlpha50
                        else
                            EFTAlpha5= 0._dl
                        end if
                    case (3) ! 3) Power law model;
                        if (deriv==0) then
                            EFTAlpha5= CP%EFTAlpha50*a**CP%EFTAlpha5Exp
                        else if (deriv==1) then
                            EFTAlpha5= CP%EFTAlpha5Exp*CP%EFTAlpha50*a**(CP%EFTAlpha5Exp-1.0_dl)
                        else if (deriv==2) then
                            EFTAlpha5= (CP%EFTAlpha5Exp-1.0_dl)*CP%EFTAlpha5Exp*CP%EFTAlpha50*&
                                &a**(CP%EFTAlpha5Exp-2.0_dl)
                        else if (deriv==3) then
                            EFTAlpha5= (CP%EFTAlpha5Exp-2.0_dl)*(CP%EFTAlpha5Exp-1.0_dl)*&
                                &CP%EFTAlpha5Exp*CP%EFTAlpha50*a**(CP%EFTAlpha5Exp-3.0_dl)
                        end if
                    case (4) ! 4) Exponential model;
                        if (deriv==0) then
                            EFTAlpha5= Exp(CP%EFTAlpha50*a**CP%EFTAlpha5Exp)-1._dl
                        else if (deriv==1) then
                            EFTAlpha5= CP%EFTAlpha5Exp*CP%EFTAlpha50*a**(CP%EFTAlpha5Exp-1.0_dl)&
                                &*Exp(CP%EFTAlpha50*a**CP%EFTAlpha5Exp)
                        else if (deriv==2) then
                            EFTAlpha5= a**(-2._dl+CP%EFTAlpha5Exp)*Exp(a**CP%EFTAlpha5Exp*CP%EFTAlpha50)*&
                                &CP%EFTAlpha50*CP%EFTAlpha5Exp*(-1._dl+CP%EFTAlpha5Exp&
                                &+a**CP%EFTAlpha5Exp*CP%EFTAlpha50*CP%EFTAlpha5Exp)
                        else if (deriv==3) then
                            EFTAlpha5= a**(-3._dl+CP%EFTAlpha5Exp)*Exp(a**CP%EFTAlpha5Exp*CP%EFTAlpha50)*&
                                &CP%EFTAlpha50*CP%EFTAlpha5Exp*(2._dl-3._dl*(1._dl+a**CP%EFTAlpha5Exp*&
                                &CP%EFTAlpha50)*CP%EFTAlpha5Exp+(1._dl+3._dl*a**CP%EFTAlpha5Exp*&
                                &CP%EFTAlpha50+a**(2._dl*CP%EFTAlpha5Exp)*CP%EFTAlpha50**2)&
                                &*CP%EFTAlpha5Exp**2)
                        end if
                    case (5) ! 5) User defined;
                        stop 'You have to define the model for EFTAlpha5 first!'
                        if (deriv==0) then
                        else if (deriv==1) then
                        else if (deriv==2) then
                        else if (deriv==3) then
                        end if
                    case default
                        stop 'Wrong choice for EFTAlpha5'
                end select
            case (2) ! Designer mapping EFT
                select case (CP%DesignerEFTModel)
                    case (1) ! Designer f(R) models
                        EFTAlpha5=0._dl
                    case (2) ! Designer mc quintessence
                        EFTAlpha5=0._dl
                    case default
                        stop 'Wrong choice for EFTAlpha5'
                end select
            case default
                stop 'Wrong choice for EFTAlpha5'
        end select

    end function EFTAlpha5

    ! ----------------------------

    function EFTAlpha6(a,deriv)
        ! EFT function Alpha_6
        implicit none
        real(dl) :: EFTAlpha6
        real(dl) , intent(in) :: a      ! scale factor
        integer  , intent(in) :: deriv  ! order of derivative wrt the scale factor

        select case (CP%EFTflag)
            case (0) ! GR
                EFTAlpha6 = 0._dl
            case (1) ! Pure EFT
                select case (CP%PureEFTmodelAlpha6)
                    case (0) ! 0) Operator neglected;
                        EFTAlpha6 = 0._dl
                    case (1) ! 1) Constant model;
                        if (deriv==0) then
                            EFTAlpha6 = CP%EFTAlpha60
                        else
                            EFTAlpha6 = 0._dl
                        end if
                    case (2) ! 2) Linear model;
                        if (deriv==0) then
                            EFTAlpha6= CP%EFTAlpha60*a
                        else if (deriv==1) then
                            EFTAlpha6= CP%EFTAlpha60
                        else
                            EFTAlpha6= 0._dl
                        end if
                    case (3) ! 3) Power law model;
                        if (deriv==0) then
                            EFTAlpha6= CP%EFTAlpha60*a**CP%EFTAlpha6Exp
                        else if (deriv==1) then
                            EFTAlpha6= CP%EFTAlpha6Exp*CP%EFTAlpha60*a**(CP%EFTAlpha6Exp-1.0_dl)
                        else if (deriv==2) then
                            EFTAlpha6= (CP%EFTAlpha6Exp-1.0_dl)*CP%EFTAlpha6Exp*CP%EFTAlpha60*&
                                &a**(CP%EFTAlpha6Exp-2.0_dl)
                        else if (deriv==3) then
                            EFTAlpha6= (CP%EFTAlpha6Exp-2.0_dl)*(CP%EFTAlpha6Exp-1.0_dl)*&
                                &CP%EFTAlpha6Exp*CP%EFTAlpha60*a**(CP%EFTAlpha6Exp-3.0_dl)
                        end if
                    case (4) ! 4) Exponential model;
                        if (deriv==0) then
                            EFTAlpha6= Exp(CP%EFTAlpha60*a**CP%EFTAlpha6Exp)-1._dl
                        else if (deriv==1) then
                            EFTAlpha6= CP%EFTAlpha6Exp*CP%EFTAlpha60*a**(CP%EFTAlpha6Exp-1.0_dl)&
                                &*Exp(CP%EFTAlpha60*a**CP%EFTAlpha6Exp)
                        else if (deriv==2) then
                            EFTAlpha6= a**(-2._dl+CP%EFTAlpha6Exp)*Exp(a**CP%EFTAlpha6Exp*CP%EFTAlpha60)*&
                                &CP%EFTAlpha60*CP%EFTAlpha6Exp*(-1._dl+CP%EFTAlpha6Exp&
                                &+a**CP%EFTAlpha6Exp*CP%EFTAlpha60*CP%EFTAlpha6Exp)
                        else if (deriv==3) then
                            EFTAlpha6= a**(-3._dl+CP%EFTAlpha6Exp)*Exp(a**CP%EFTAlpha6Exp*CP%EFTAlpha60)*&
                                &CP%EFTAlpha60*CP%EFTAlpha6Exp*(2._dl-3._dl*(1._dl+a**CP%EFTAlpha6Exp*&
                                &CP%EFTAlpha60)*CP%EFTAlpha6Exp+(1._dl+3._dl*a**CP%EFTAlpha6Exp*&
                                &CP%EFTAlpha60+a**(2._dl*CP%EFTAlpha6Exp)*CP%EFTAlpha60**2)&
                                &*CP%EFTAlpha6Exp**2)
                        end if
                    case (5) ! 5) User defined;
                        stop 'You have to define the model for EFTAlpha6 first!'
                        if (deriv==0) then
                        else if (deriv==1) then
                        else if (deriv==2) then
                        else if (deriv==3) then
                        end if
                    case default
                        stop 'Wrong choice for EFTAlpha6'
                end select
            case (2) ! Designer mapping EFT
                select case (CP%DesignerEFTModel)
                    case (1) ! Designer f(R) models
                        EFTAlpha6=0._dl
                    case (2) ! Designer mc quintessence
                        EFTAlpha6=0._dl
                    case default
                        stop 'Wrong choice for EFTAlpha6'
                end select
            case default
                stop 'Wrong choice for EFTAlpha6'
        end select

    end function EFTAlpha6

    ! ----------------------------

    function EFTLambdaTemp(a,deriv)
        ! EFT function Lambda
        ! This function is defined as Lambda/Mpl*a^2
        implicit none
        real(dl):: EFTLambdaTemp
        real(dl) , intent(in) :: a      ! scale factor
        integer  , intent(in) :: deriv  ! order of derivative wrt the scale factor

        select case (CP%EFTflag)
            case (0) ! GR
                EFTLambdaTemp = 0._dl
            case (1) ! Pure EFT: Lambda is obtained through EFT designer approach (in equations.f90).
                EFTLambdaTemp = 0._dl
            case (2) ! Designer mapping EFT
                select case (CP%DesignerEFTModel)
                    case (1) ! Designer f(R) models
                        if (deriv==0) then
                            EFTLambdaTemp = Designed_EFT_Function(EFT_Lambda_Des, a, 0._dl)
                        else if (deriv==1) then
                            EFTLambdaTemp = Designed_EFT_Function(EFT_Lambdadot_Des, a, 0._dl)
                        end if
                    case (2) ! Designer mc quintessence: Lambda is obtained through EFT designer approach (in equations.f90).
                        EFTLambdaTemp = 0._dl
                    case default
                        stop 'Wrong choice for Lambda'
                end select
            case default
                stop 'Wrong choice for Lambda'
        end select

    end function EFTLambdaTemp

    ! ----------------------------

    function EFTcTemp(a,deriv)
        ! EFT function c
        ! This function is defined as c/Mpl*a^2
        implicit none
        real(dl):: EFTcTemp
        real(dl) , intent(in) :: a      ! scale factor
        integer  , intent(in) :: deriv  ! order of derivative wrt the scale factor

        select case (CP%EFTflag)
            case (0) ! GR
                EFTcTemp = 0._dl
            case (1) ! Pure EFT: c is obtained through EFT designer approach (in equations.f90).
                EFTcTemp = 0._dl
            case (2) ! Designer mapping EFT
                select case (CP%DesignerEFTModel)
                    case (1) ! Designer f(R) models
                        EFTcTemp = 0._dl
                    case (2) ! Designer mc quintessence: c is obtained through EFT designer approach (in equations.f90).
                        EFTcTemp = 0._dl
                    case default
                        stop 'Wrong choice for c'
                end select
            case default
                stop 'Wrong choice for c'
        end select

    end function EFTcTemp

end module EFTfunctions

! -------------------------------------------------------------------------------------------------

!Code used to detect the return to GR (RGR) of the considered model.
module EFTreturntoGR
    use EFTDef
    use EFTfunctions
    use MassiveNu
    implicit none

    !We need to define this to interface the EFT functions with the RGR code.
    abstract interface
        function EFTfunction(a,ind) result(y)
            use precision
            implicit none
            real(dl), intent(in) :: a
            integer , intent(in) :: ind
            real(dl)  y
        end function EFTfunction
    end interface
    !This two variables stores the pointer to the specific EFT function which is considered
    !by the RGR code.
    integer :: deriv
    procedure(EFTfunction), pointer :: EFTfuncTemp

contains

    ! ----------------------------

    subroutine EFTCheckReturnToGR
        ! RGR test main subroutine.
        implicit none

        real(dl) :: OmegaTime=0._dl, OmegaDotTime=0._dl
        real(dl) :: EFTcTime=0._dl, EFTcDotTime=0._dl, EFTLambdaTime=0._dl, EFTLambdaDotTime=0._dl
        real(dl) :: EFTAlpha1Time=0._dl, EFTAlpha1DotTime=0._dl
        real(dl) :: EFTAlpha2Time=0._dl, EFTAlpha2DotTime=0._dl
        real(dl) :: EFTAlpha3Time=0._dl, EFTAlpha3DotTime=0._dl
        real(dl) :: EFTAlpha4Time=0._dl, EFTAlpha4DotTime=0._dl
        real(dl) :: EFTAlpha5Time=0._dl, EFTAlpha5DotTime=0._dl
        real(dl) :: EFTAlpha6Time=0._dl, EFTAlpha6DotTime=0._dl

        real(dl) :: temp

        if (Feedbacklevel>1) write(*,*) 'EFTCAMB: checking return to GR'

        ! 1) Test Omega:
        deriv = 0
        EFTfuncTemp => EFTOmega
        call EFTdetectionRGR(EFTfunctionRGR, EFTturnonpiInitial, 1._dl, OmegaTime)
        ! 2) Test Omega Dot:
        call EFTdetectionRGR(RGR_EFTOmegaDot, EFTturnonpiInitial, 1._dl, OmegaDotTime)
        ! 3) Test c:
        call EFTdetectionRGR(RGR_EFTc, EFTturnonpiInitial, 1._dl, EFTcTime)
        ! 4) Test c Dot:
        call EFTdetectionRGR(RGR_EFTcDot, EFTturnonpiInitial, 1._dl, EFTcDotTime)
        ! 5) Test Lambda:
        call EFTdetectionRGR(RGR_EFTLambda, EFTturnonpiInitial, 1._dl, EFTLambdaTime)
        ! 6) Test Lambda Dot:
        call EFTdetectionRGR(RGR_EFTLambdaDot, EFTturnonpiInitial, 1._dl, EFTLambdaDotTime)
        ! 7) Test Alpha1:
        deriv = 0
        EFTfuncTemp => EFTAlpha1
        call EFTdetectionRGR(EFTfunctionRGR, EFTturnonpiInitial, 1._dl, EFTAlpha1Time)
        ! 8) Test Alpha2:
        deriv = 0
        EFTfuncTemp => EFTAlpha2
        call EFTdetectionRGR(EFTfunctionRGR, EFTturnonpiInitial, 1._dl, EFTAlpha2Time)
        ! 9) Test Alpha3:
        deriv = 0
        EFTfuncTemp => EFTAlpha3
        call EFTdetectionRGR(EFTfunctionRGR, EFTturnonpiInitial, 1._dl, EFTAlpha3Time)
        ! 10) Test Alpha4:
        deriv = 0
        EFTfuncTemp => EFTAlpha4
        call EFTdetectionRGR(EFTfunctionRGR, EFTturnonpiInitial, 1._dl, EFTAlpha4Time)
        ! 11) Test Alpha5:
        deriv = 0
        EFTfuncTemp => EFTAlpha5
        call EFTdetectionRGR(EFTfunctionRGR, EFTturnonpiInitial, 1._dl, EFTAlpha5Time)
        ! 12) Test Alpha6:
        deriv = 0
        EFTfuncTemp => EFTAlpha6
        call EFTdetectionRGR(EFTfunctionRGR, EFTturnonpiInitial, 1._dl, EFTAlpha6Time)
        ! 13) Test Alpha1 prime:
        deriv = 1
        EFTfuncTemp => EFTAlpha1
        call EFTdetectionRGR(EFTfunctionRGR, EFTturnonpiInitial, 1._dl, EFTAlpha1DotTime)
        ! 14) Test Alpha2 prime:
        deriv = 1
        EFTfuncTemp => EFTAlpha2
        call EFTdetectionRGR(EFTfunctionRGR, EFTturnonpiInitial, 1._dl, EFTAlpha2DotTime)
        ! 15) Test Alpha3 prime:
        deriv = 1
        EFTfuncTemp => EFTAlpha3
        call EFTdetectionRGR(EFTfunctionRGR, EFTturnonpiInitial, 1._dl, EFTAlpha3DotTime)
        ! 16) Test Alpha4 prime:
        deriv = 1
        EFTfuncTemp => EFTAlpha4
        call EFTdetectionRGR(EFTfunctionRGR, EFTturnonpiInitial, 1._dl, EFTAlpha4DotTime)
        ! 17) Test Alpha5 prime:
        deriv = 1
        EFTfuncTemp => EFTAlpha5
        call EFTdetectionRGR(EFTfunctionRGR, EFTturnonpiInitial, 1._dl, EFTAlpha5DotTime)
        ! 18) Test Alpha6 prime:
        deriv = 1
        EFTfuncTemp => EFTAlpha6
        call EFTdetectionRGR(EFTfunctionRGR, EFTturnonpiInitial, 1._dl, EFTAlpha6DotTime)

        ! 19) Now determine EFTturnonpi

        temp = min(OmegaTime,OmegaDotTime, EFTcTime, EFTcDotTime, EFTLambdaTime, EFTLambdaDotTime,&
            &EFTAlpha1Time, EFTAlpha1DotTime, EFTAlpha2Time, EFTAlpha2DotTime,&
            &EFTAlpha3Time, EFTAlpha3DotTime, EFTAlpha4Time, EFTAlpha4DotTime,&
            &EFTAlpha5Time, EFTAlpha5DotTime, EFTAlpha6Time, EFTAlpha6DotTime)

        if (Feedbacklevel>2) then
            write(*,*) 'EFTCAMB: RGR times'
            write(*,*) OmegaTime,OmegaDotTime, EFTcTime, EFTcDotTime, EFTLambdaTime, EFTLambdaDotTime,&
                &EFTAlpha1Time, EFTAlpha1DotTime, EFTAlpha2Time, EFTAlpha2DotTime,&
                &EFTAlpha3Time, EFTAlpha3DotTime, EFTAlpha4Time, EFTAlpha4DotTime,&
                &EFTAlpha5Time, EFTAlpha5DotTime, EFTAlpha6Time, EFTAlpha6DotTime
        end if

        if (temp.le.EFTturnonpiInitial) then
            EFTturnonpi = EFTturnonpiInitial
        else if (temp>EFTturnonpiInitial) then
            EFTturnonpi = temp
        end if

        if (Feedbacklevel>1) write(*,*) 'RGR: a=', EFTturnonpi

    end subroutine EFTCheckReturnToGR

    ! ----------------------------

    function EFTfunctionRGR(a)
        ! This function is needed to make the two input variable EFT functions into a single input functions.
        implicit none
        real(dl), intent(in) :: a
        real(dl) :: EFTfunctionRGR
        EFTfunctionRGR = ABS(EFTfuncTemp(a,deriv))
    end function EFTfunctionRGR

    ! ----------------------------

    subroutine EFTdetectionRGR(EFTfunction,TimeStart, TimeEnd, RGRtime)
        ! Subroutine that test the return to GR of the EFT functions.
        implicit none
        external EFTfunction
        real(dl) :: EFTfunction
        real(dl), intent(in)  :: TimeStart, TimeEnd
        real(dl), intent(out) :: RGRtime
        real(dl) :: zbrent
        external zbrent

        real(dl) :: temp1, temp2, temp3
        integer  :: j
        logical :: succes

        temp1 = EFTfunction(TimeStart) - EFTtoGR
        temp2 = EFTfunction(TimeEnd) - EFTtoGR

        if (temp1*temp2>0.and.temp2>0) then         ! always not GR
            RGRtime = TimeStart
            return
        else if (temp1*temp2>0.and.temp2<0) then    ! Omega is always GR
            RGRtime = 1.1_dl*TimeEnd
            return
        else if (temp1*temp2<0 .and. temp2<0) then  ! Omega is not GR at early times
            RGRtime = TimeStart
            return
        else if (temp1*temp2<0.and. temp2>0) then   ! There is a transition
            RGRtime = zbrent(EFTfunction,TimeStart,TimeEnd,1.d-10,EFTtoGR,succes)
            if (.not.succes) RGRtime = TimeStart
            return
        end if

        RGRtime = TimeStart
        return

    end subroutine EFTdetectionRGR

    ! ----------------------------

    function RGR_EFTOmegaDot(a)
        !Dummy function for the absolute value of EFTOmegaDot
        implicit none
        real(dl) a,RGR_EFTOmegaDot, EFT_H0
        real(dl) a2, grhob_t,grhoc_t,grhor_t,grhog_t,grhov_t,grho,adotoa
        real(dl) EFT_grhonu, EFT_gpinu, grhormass_t, EFT_grhonu_tot                 
        integer  nu_i
        a2 = a*a
        grhob_t=grhob/a
        grhoc_t=grhoc/a
        grhor_t=grhornomass/a2
        grhog_t=grhog/a2
        grhov_t=grhov*EFTw(a,3)
        grho=grhob_t+grhoc_t+grhor_t+grhog_t+grhov_t
        EFT_grhonu_tot = 0._dl
        if (CP%Num_Nu_Massive /= 0) then                                            
            do nu_i = 1, CP%Nu_mass_eigenstates
                EFT_grhonu    = 0._dl
                EFT_gpinu     = 0._dl
                grhormass_t=grhormass(nu_i)/a**2
                call Nu_background(a*nu_masses(nu_i),EFT_grhonu,EFT_gpinu)
                EFT_grhonu_tot = EFT_grhonu_tot + grhormass_t*EFT_grhonu
            end do
        end if
        grho  = grho  + EFT_grhonu_tot
        adotoa=sqrt(grho/3)
        EFT_H0 = (CP%h0/c_EFT)*1000._dl
        RGR_EFTOmegaDot = ABS(a*adotoa*EFTOmega(a,1))
        return
    end function RGR_EFTOmegaDot

    ! ----------------------------

    function RGR_EFTc(a)
        !Dummy function for the absolute value of EFTc
        implicit none
        real(dl) a,RGR_EFTc
        real(dl) a2, grhob_t,grhoc_t,grhor_t,grhog_t,grhov_t,grho, gpres, EFT_H0
        real(dl) adotoa, adotdota, Hdot
        real(dl) EFTOmegaV, EFTOmegaP, EFTOmegaPP
        real(dl) EFTc
        real(dl) EFT_grhonu, EFT_gpinu, EFT_grhonu_tot, EFT_gpinu_tot, grhormass_t    
        integer  nu_i
        a2 = a*a
        grhob_t=grhob/a
        grhoc_t=grhoc/a
        grhor_t=grhornomass/a2
        grhog_t=grhog/a2
        grhov_t=grhov*EFTw(a,3)
        grho=grhob_t+grhoc_t+grhor_t+grhog_t+grhov_t
        gpres= (grhog_t+grhor_t)/3._dl +EFTw(a,0)*grhov_t
        EFT_gpinu_tot = 0._dl
        EFT_grhonu_tot = 0._dl
        if (CP%Num_Nu_Massive /= 0) then                                            
            do nu_i = 1, CP%Nu_mass_eigenstates
                EFT_grhonu    = 0._dl
                EFT_gpinu     = 0._dl
                grhormass_t=grhormass(nu_i)/a**2
                call Nu_background(a*nu_masses(nu_i),EFT_grhonu,EFT_gpinu)
                EFT_grhonu_tot = EFT_grhonu_tot + grhormass_t*EFT_grhonu
                EFT_gpinu_tot  = EFT_gpinu_tot + grhormass_t*EFT_gpinu
            end do
        end if

        grho  = grho  + EFT_grhonu_tot
        gpres = gpres + EFT_gpinu_tot
        
        EFT_H0 = (CP%h0/c_EFT)*1000._dl
        ! 3) Hubble and derivatives:
        adotoa=sqrt(grho/3)
        adotdota=(adotoa*adotoa-gpres)/2._dl
        Hdot =adotdota-adotoa**2._dl
        ! 4) EFT functions:
        EFTOmegaV = EFTOmega(a,0)
        EFTOmegaP = EFTOmega(a,1)
        EFTOmegaPP = EFTOmega(a,2)

        select case (CP%EFTflag) ! 8*pi*G*c*a^2
            case (1) ! Pure EFT
                EFTc = (adotoa*adotoa - Hdot)*(EFTOmegaV + a*EFTOmegaP*0.5_dl) &
                    & - 0.5_dl*a2*adotoa*adotoa*EFTOmegaPP&
                    & + 0.5_dl*grhov_t*(1._dl+EFTw(a,0))
            case (2) ! Designer mapping
                select case (CP%DesignerEFTModel)
                    case (2) ! Designer mc quintessence
                        EFTc = (adotoa*adotoa - Hdot)*(EFTOmegaV + a*EFTOmegaP*0.5_dl) &
                            & - 0.5_dl*a2*adotoa*adotoa*EFTOmegaPP&
                            & + 0.5_dl*grhov_t*(1._dl+EFTw(a,0))
                    case default ! normally we can just use the designed one.
                        EFTc = EFTcTemp(a,0)
                end select
        end select

        RGR_EFTc = ABS(EFTc/a2)
        return
    end function RGR_EFTc

    ! ----------------------------

    function RGR_EFTcDot(a)
        ! Dummy function for the absolute value of EFTcdot
        implicit none
        real(dl) a,RGR_EFTcDot
        real(dl) a2, grhob_t,grhoc_t,grhor_t,grhog_t,grhov_t,grho, gpres, EFT_H0
        real(dl) adotoa, adotdota, Hdot,Hdotdot
        real(dl) EFTOmegaV, EFTOmegaP, EFTOmegaPP,EFTOmegaPPP
        real(dl) EFTc, EFTcdot, EFTLambda, EFTLambdadot
        real(dl) EFT_grhonu, EFT_gpinu, EFT_grhonu_tot, EFT_gpinu_tot, EFT_gpinudot_tot, grhormass_t    
        integer  nu_i

        a2 = a*a
        grhob_t=grhob/a
        grhoc_t=grhoc/a
        grhor_t=grhornomass/a2
        grhog_t=grhog/a2
        grhov_t=grhov*EFTw(a,3)
        grho=grhob_t+grhoc_t+grhor_t+grhog_t+grhov_t
        gpres= (grhog_t+grhor_t)/3._dl +EFTw(a,0)*grhov_t
        EFT_H0 = (CP%h0/c_EFT)*1000._dl
        
        EFT_grhonu_tot = 0._dl
        EFT_gpinu_tot = 0._dl
        if (CP%Num_Nu_Massive /= 0) then                                            
            do nu_i = 1, CP%Nu_mass_eigenstates
                EFT_grhonu    = 0._dl
                EFT_gpinu     = 0._dl
                grhormass_t=grhormass(nu_i)/a**2
                call Nu_background(a*nu_masses(nu_i),EFT_grhonu,EFT_gpinu)
                EFT_grhonu_tot = EFT_grhonu_tot + grhormass_t*EFT_grhonu
                EFT_gpinu_tot  = EFT_gpinu_tot + grhormass_t*EFT_gpinu
            end do
        end if

        grho  = grho  + EFT_grhonu_tot
        gpres = gpres + EFT_gpinu_tot
        ! 3) Hubble and derivatives:
        adotoa=sqrt(grho/3)
        adotdota=(adotoa*adotoa-gpres)/2._dl
        Hdot =adotdota-adotoa**2._dl

        EFT_gpinu_tot = 0._dl
        EFT_grhonu_tot = 0._dl
        EFT_gpinudot_tot = 0._dl

        if (CP%Num_Nu_Massive /= 0) then                                            
            do nu_i = 1, CP%Nu_mass_eigenstates
                EFT_grhonu    = 0._dl
                EFT_gpinu     = 0._dl
                grhormass_t=grhormass(nu_i)/a**2
                call Nu_background(a*nu_masses(nu_i),EFT_grhonu,EFT_gpinu)
                EFT_grhonu_tot = EFT_grhonu_tot + grhormass_t*EFT_grhonu
                EFT_gpinu_tot  = EFT_gpinu_tot + grhormass_t*EFT_gpinu
                EFT_gpinudot_tot  = EFT_gpinudot_tot + grhormass_t*(Nu_pidot(a*nu_masses(nu_i),adotoa, EFT_gpinu)&
                    & -4._dl*adotoa*EFT_gpinu)
            end do
        end if

        Hdotdot = 2._dl*adotoa*Hdot &
            & + 0.5_dl*adotoa*(grhob_t + grhoc_t + 8._dl*(grhog_t+grhor_t)/3._dl)&
            & + 0.5_dl*adotoa*grhov_t*((1._dl+EFTw(a,0))*(1._dl+3._dl*EFTw(a,0))-a*EFTw(a,1))&
            & + adotoa/6._dl*EFT_grhonu_tot -0.5_dl*adotoa*EFT_gpinu_tot -0.5_dl*EFT_gpinudot_tot   

        ! 4) EFT functions:
        EFTOmegaV = EFTOmega(a,0)
        EFTOmegaP = EFTOmega(a,1)
        EFTOmegaPP = EFTOmega(a,2)
        EFTOmegaPPP = EFTOmega(a,3)

        select case (CP%EFTflag) ! 8*pi*G*cdot*a^2
            case (1) ! Pure EFT
                EFTcdot = -EFTOmegaV*(Hdotdot-4._dl*adotoa*Hdot+2._dl*adotoa*adotoa*adotoa) &
                    & + 0.5_dl*a*EFTOmegaP*(-Hdotdot+adotoa*Hdot+adotoa*adotoa*adotoa)&
                    & +0.5_dl*a2*adotoa*EFTOmegaPP*(adotoa*adotoa-3._dl*Hdot)&
                    & -0.5_dl*a*a2*adotoa*adotoa*adotoa*EFTOmegaPPP&
                    & +0.5_dl*adotoa*grhov_t*(-3._dl*(1._dl+EFTw(a,0))**2 + a*EFTw(a,1))
            case (2) ! Designer mapping
                select case (CP%DesignerEFTModel)
                    case (2) ! Designer mc quintessence
                        EFTcdot = -EFTOmegaV*(Hdotdot-4._dl*adotoa*Hdot+2._dl*adotoa*adotoa*adotoa) &
                            & + 0.5_dl*a*EFTOmegaP*(-Hdotdot+adotoa*Hdot+adotoa*adotoa*adotoa)&
                            & +0.5_dl*a2*adotoa*EFTOmegaPP*(adotoa*adotoa-3._dl*Hdot)&
                            & -0.5_dl*a*a2*adotoa*adotoa*adotoa*EFTOmegaPPP&
                            & +0.5_dl*adotoa*grhov_t*(-3._dl*(1._dl+EFTw(a,0))**2 + a*EFTw(a,1))
                    case default
                        EFTcdot = EFTcTemp(a,1)
                end select
        end select

        RGR_EFTcDot = ABS(EFTcdot/a2)
        return
    end function RGR_EFTcDOT

    ! ----------------------------

    function RGR_EFTLambda(a)
        ! Dummy function for the absolute value of EFTLambda
        implicit none
        real(dl) a,RGR_EFTLambda
        real(dl) a2, grhob_t,grhoc_t,grhor_t,grhog_t,grhov_t,grho, gpres, EFT_H0
        real(dl) adotoa, adotdota, Hdot,Hdotdot
        real(dl) EFTOmegaV, EFTOmegaP, EFTOmegaPP,EFTOmegaPPP
        real(dl)  EFTc, EFTcdot, EFTLambda, EFTLambdadot
        real(dl) EFT_grhonu, EFT_gpinu, grhormass_t    
        real(dl) EFT_grhonu_tot, EFT_gpinu_tot, EFT_gpinudot_tot 
        integer  nu_i        

        a2 = a*a
        grhob_t=grhob/a
        grhoc_t=grhoc/a
        grhor_t=grhornomass/a2
        grhog_t=grhog/a2
        grhov_t=grhov*EFTw(a,3)
        grho=grhob_t+grhoc_t+grhor_t+grhog_t+grhov_t
        gpres= (grhog_t+grhor_t)/3._dl +EFTw(a,0)*grhov_t
        EFT_H0 = (CP%h0/c_EFT)*1000._dl

        EFT_gpinu_tot = 0._dl
        EFT_grhonu_tot = 0._dl

        if (CP%Num_Nu_Massive /= 0) then                                            
            do nu_i = 1, CP%Nu_mass_eigenstates
                EFT_grhonu    = 0._dl
                EFT_gpinu     = 0._dl
                grhormass_t=grhormass(nu_i)/a**2
                call Nu_background(a*nu_masses(nu_i),EFT_grhonu,EFT_gpinu)
                EFT_grhonu_tot = EFT_grhonu_tot + grhormass_t*EFT_grhonu
                EFT_gpinu_tot  = EFT_gpinu_tot + grhormass_t*EFT_gpinu
            end do
        end if

        grho  = grho  + EFT_grhonu_tot
        gpres = gpres + EFT_gpinu_tot
        ! 3) Hubble and derivatives:
        adotoa=sqrt(grho/3)
        adotdota=(adotoa*adotoa-gpres)/2._dl
        Hdot =adotdota-adotoa**2._dl

        EFT_gpinu_tot = 0._dl
        EFT_grhonu_tot = 0._dl
        EFT_gpinudot_tot = 0._dl

        if (CP%Num_Nu_Massive /= 0) then                                            
            do nu_i = 1, CP%Nu_mass_eigenstates
                EFT_grhonu    = 0._dl
                EFT_gpinu     = 0._dl
                grhormass_t=grhormass(nu_i)/a**2
                call Nu_background(a*nu_masses(nu_i),EFT_grhonu,EFT_gpinu)
                EFT_grhonu_tot = EFT_grhonu_tot + grhormass_t*EFT_grhonu
                EFT_gpinu_tot  = EFT_gpinu_tot + grhormass_t*EFT_gpinu
                EFT_gpinudot_tot  = EFT_gpinudot_tot + grhormass_t*(Nu_pidot(a*nu_masses(nu_i),adotoa, EFT_gpinu)&
                    & -4._dl*adotoa*EFT_gpinu)
            end do
        end if

        Hdotdot = 2._dl*adotoa*Hdot &
            & + 0.5_dl*adotoa*(grhob_t + grhoc_t + 8._dl*(grhog_t+grhor_t)/3._dl)&
            & + 0.5_dl*adotoa*grhov_t*((1._dl+EFTw(a,0))*(1._dl+3._dl*EFTw(a,0))-a*EFTw(a,1))&
            & + adotoa/6._dl*EFT_grhonu_tot -0.5_dl*adotoa*EFT_gpinu_tot -0.5_dl*EFT_gpinudot_tot   
        ! 4) EFT functions:
        EFTOmegaV = EFTOmega(a,0)
        EFTOmegaP = EFTOmega(a,1)
        EFTOmegaPP = EFTOmega(a,2)
        EFTOmegaPPP = EFTOmega(a,3)

        select case (CP%EFTflag) !8*pi*G*Lambda*a^2
            case (1) ! Pure EFT
                EFTLambda = +EFTw(a,0)*grhov_t &
                    &-EFTOmegaV*(2._dl*Hdot+adotoa**2._dl) &
                    &-a*EFTOmegaP*(2._dl*adotoa**2._dl + Hdot) &
                    &-a2*adotoa**2._dl*EFTOmegaPP
            case (2) ! Designer mapping
                select case (CP%DesignerEFTModel)
                    case (2) ! Designer mc quintessence
                        EFTLambda = +EFTw(a,0)*grhov_t &
                            &-EFTOmegaV*(2._dl*Hdot+adotoa**2._dl) &
                            &-a*EFTOmegaP*(2._dl*adotoa**2._dl + Hdot) &
                            &-a2*adotoa**2._dl*EFTOmegaPP
                    case default
                        EFTLambda = EFTLambdaTemp(a,0)
                end select
        end select

        RGR_EFTLambda = ABS(EFTLambda/a2 + grhov)
        return
    end function RGR_EFTLambda

    ! ----------------------------

    function RGR_EFTLambdaDot(a)
        ! Dummy function for the absolute value of EFTLambdaDot
        implicit none
        real(dl) a,RGR_EFTLambdaDot
        real(dl) a2, grhob_t,grhoc_t,grhor_t,grhog_t,grhov_t,grho, gpres, EFT_H0
        real(dl) adotoa, adotdota, Hdot,Hdotdot
        real(dl) EFTOmegaV, EFTOmegaP, EFTOmegaPP,EFTOmegaPPP
        real(dl) EFTc, EFTcdot, EFTLambda, EFTLambdadot
        real(dl) EFT_grhonu, EFT_gpinu, grhormass_t    
        real(dl) EFT_grhonu_tot, EFT_gpinu_tot, EFT_gpinudot_tot 
        integer  nu_i  

        a2 = a*a
        grhob_t=grhob/a
        grhoc_t=grhoc/a
        grhor_t=grhornomass/a2
        grhog_t=grhog/a2
        grhov_t=grhov*EFTw(a,3)
        grho=grhob_t+grhoc_t+grhor_t+grhog_t+grhov_t
        gpres= (grhog_t+grhor_t)/3._dl +EFTw(a,0)*grhov_t
        EFT_H0 = (CP%h0/c_EFT)*1000._dl

        if (CP%Num_Nu_Massive /= 0) then                                            
            do nu_i = 1, CP%Nu_mass_eigenstates
                EFT_grhonu    = 0._dl
                EFT_gpinu     = 0._dl
                grhormass_t=grhormass(nu_i)/a**2
                call Nu_background(a*nu_masses(nu_i),EFT_grhonu,EFT_gpinu)
                EFT_grhonu_tot = EFT_grhonu_tot + grhormass_t*EFT_grhonu
                EFT_gpinu_tot  = EFT_gpinu_tot + grhormass_t*EFT_gpinu
            end do
        end if

        grho  = grho  + EFT_grhonu_tot
        gpres = gpres + EFT_gpinu_tot
        ! 3) Hubble and derivatives:
        adotoa=sqrt(grho/3)
        adotdota=(adotoa*adotoa-gpres)/2._dl
        Hdot =adotdota-adotoa**2._dl

        EFT_gpinu_tot = 0._dl
        EFT_grhonu_tot = 0._dl
        EFT_gpinudot_tot = 0._dl

        if (CP%Num_Nu_Massive /= 0) then                                            
            do nu_i = 1, CP%Nu_mass_eigenstates
                EFT_grhonu    = 0._dl
                EFT_gpinu     = 0._dl
                grhormass_t=grhormass(nu_i)/a**2
                call Nu_background(a*nu_masses(nu_i),EFT_grhonu,EFT_gpinu)
                EFT_grhonu_tot = EFT_grhonu_tot + grhormass_t*EFT_grhonu
                EFT_gpinu_tot  = EFT_gpinu_tot + grhormass_t*EFT_gpinu
                EFT_gpinudot_tot  = EFT_gpinudot_tot + grhormass_t*(Nu_pidot(a*nu_masses(nu_i),adotoa, EFT_gpinu)&
                    & -4._dl*adotoa*EFT_gpinu)
            end do
        end if

        Hdotdot = 2._dl*adotoa*Hdot &
            & + 0.5_dl*adotoa*(grhob_t + grhoc_t + 8._dl*(grhog_t+grhor_t)/3._dl)&
            & + 0.5_dl*adotoa*grhov_t*((1._dl+EFTw(a,0))*(1._dl+3._dl*EFTw(a,0))-a*EFTw(a,1))&
            & + adotoa/6._dl*EFT_grhonu_tot -0.5_dl*adotoa*EFT_gpinu_tot -0.5_dl*EFT_gpinudot_tot   
        ! 4) EFT functions:
        EFTOmegaV = EFTOmega(a,0)
        EFTOmegaP = EFTOmega(a,1)
        EFTOmegaPP = EFTOmega(a,2)
        EFTOmegaPPP = EFTOmega(a,3)

        select case (CP%EFTflag) ! 8*pi*G*Ldot*a^2
            case (1) ! Pure EFT
                EFTLambdadot = -2._dl*EFTOmegaV*(Hdotdot-adotoa*Hdot-adotoa*adotoa*adotoa)&
                    & - a*EFTOmegaP*(4._dl*adotoa*Hdot+Hdotdot)&
                    & -a2*EFTOmegaPP*adotoa*(3._dl*Hdot+2._dl*adotoa*adotoa)&
                    & -a*a2*EFTOmegaPPP*adotoa*adotoa*adotoa&
                    & +grhov_t*adotoa*(a*EFTw(a,1)-3._dl*EFTw(a,0)*(1._dl+EFTw(a,0)))
            case (2) ! Designer mapping
                select case (CP%DesignerEFTModel)
                    case (2) ! Designer mc quintessence
                        EFTLambdadot = -2._dl*EFTOmegaV*(Hdotdot-adotoa*Hdot-adotoa*adotoa*adotoa)&
                            & - a*EFTOmegaP*(4._dl*adotoa*Hdot+Hdotdot)&
                            & -a2*EFTOmegaPP*adotoa*(3._dl*Hdot+2._dl*adotoa*adotoa)&
                            & -a*a2*EFTOmegaPPP*adotoa*adotoa*adotoa&
                            & +grhov_t*adotoa*(a*EFTw(a,1)-3._dl*EFTw(a,0)*(1._dl+EFTw(a,0)))
                    case default
                        EFTLambdadot = EFTLambdaTemp(a,1)
                end select
        end select

        RGR_EFTLambdaDot = ABS(EFTLambdadot/a2)
        return
    end function RGR_EFTLambdaDot

end module EFTreturntoGR

! -------------------------------------------------------------------------------------------------

!Performs the stability check for the considered model.
module EFTstability
    use EFTDef
    use EFTfunctions
    use MassiveNu
    implicit none
contains

    ! ----------------------------

    subroutine EFTCheck_Stability(success, astart, aend)
        ! CheckStability subroutine.
        ! Three scanning strategies: linear sampling from astart to aend and two log samplig on the borders.
        implicit none

        logical  :: success
        real(dl), intent(in) :: astart, aend

        real(dl) :: Atest, y
        real(dl) :: LogSamplingScale
        integer  :: ind=1
        integer  :: indMax = 10000 ! Number of points sampled.

        if (Feedbacklevel>1) write(*,*) 'EFTCAMB: checking stability of the theory'

        success = .true.
        LogSamplingScale = -10._dl

        ! 1) Do linear sampling from astart to aend.
        do ind=1, indMax
            Atest = astart + REAL(ind-1)*(aend-astart)/REAL(indMax-1)
            if (.not.EFTStabilityComputation(Atest)) then
                success = .false.
                return
            end if
        end do
        ! 2) Do Log sampling close to astart
        do ind=1, indMax
            y = LogSamplingScale + REAL(ind-1)*(0._dl-LogSamplingScale)/REAL(indMax-1)
            Atest = astart +(aend-astart)*10._dl**y
            if (.not.EFTStabilityComputation(Atest)) then
                success = .false.
                return
            end if
        end do
        ! 3) Do Log sampling close to aend
        do ind=1, indMax
            y = LogSamplingScale + REAL(ind-1)*(0._dl-LogSamplingScale)/REAL(indMax-1)
            Atest = aend +(astart-aend)*10._dl**y
            if (.not.EFTStabilityComputation(Atest)) then
                success = .false.
                return
            end if
        end do

        return
    end subroutine EFTCheck_Stability

    ! ----------------------------

    function EFTStabilityComputation(a)
        !This function computes if the stability requirements are fullfilled in a given time.
        implicit none
        ! 1) Definitions of variables:
        logical :: EFTStabilityComputation
        real(dl), intent(in) :: a
        real(dl) :: a2,k,k2,grhob_t,grhor_t,grhoc_t,grhog_t,grhov_t, grho, gpres, EFT_H0
        real(dl) :: adotoa, adotdota, Hdot, Hdotdot
        real(dl) :: EFTOmegaV, EFTOmegaP,EFTOmegaPP,EFTOmegaPPP, EFTc, EFTLambda, EFTcdot, EFTLambdadot
        real(dl) :: EFTAlpha1V, EFTAlpha1P, EFTAlpha2V, EFTAlpha2P, EFTAlpha3V, EFTAlpha3P
        real(dl) :: EFTAlpha4V, EFTAlpha4P, EFTAlpha5V, EFTAlpha5P, EFTAlpha6V, EFTAlpha6P
        real(dl) :: EFTgrhoq,EFTgpresq,EFTgrhodotq,EFTgpresdotq
        real(dl) :: EFTpiA1, EFTpiA2, EFTpiB, EFTpiC, EFTpiD, EFTpiD1, EFTpiD2, EFTAT
        real(dl) :: EFTtemp_H0,  EFTtemp_Omegac, EFTtemp_Omegab, EFTtemp_Omegav, EFTtemp_TCMB, EFTtemp_nu_massless_degeneracy
        real(dl) :: EFTtemp_grhom, EFTtemp_grhog, EFTtemp_grhor
        real(dl) :: EFT_grhonu, EFT_gpinu, grhormass_t    
        real(dl) :: EFT_grhonu_tot, EFT_gpinu_tot, EFT_gpinudot_tot 
        integer  :: nu_i
        ! 2) Computation of various quantities:
        EFTtemp_H0 = CP%H0
        EFTtemp_Omegac = CP%omegac
        EFTtemp_Omegab = CP%omegab
        EFTtemp_Omegav = CP%omegav
        EFTtemp_TCMB = CP%tcmb
        EFTtemp_nu_massless_degeneracy = CP%Num_Nu_massless
        EFTtemp_grhom = 3*EFTtemp_H0**2/c_EFT**2*1000**2
        EFTtemp_grhog = kappa_EFT/c_EFT**2*4*sigma_boltz_EFT/c_EFT**3*EFTtemp_TCMB**4*Mpc_EFT**2 !8*pi*G/c^2*4*sigma_B/c^3 T^4
        EFTtemp_grhor = 7._dl/8*(4._dl/11)**(4._dl/3)*EFTtemp_grhog
        ! 2) Computation of various quantities:
        a2=a*a
        grhob_t=EFTtemp_grhom*EFTtemp_Omegab/a
        grhoc_t=EFTtemp_grhom*EFTtemp_Omegac/a
        grhor_t=EFTtemp_grhor*EFTtemp_nu_massless_degeneracy/a2
        grhog_t=EFTtemp_grhog/a2
        grhov_t=EFTtemp_grhom*EFTtemp_Omegav*EFTw(a,3)
        grho=grhob_t+grhoc_t+grhor_t+grhog_t+grhov_t
        gpres= (grhog_t+grhor_t)/3._dl +EFTw(a,0)*grhov_t
        EFT_H0 = (CP%h0/c_EFT)*1000._dl

        EFT_grhonu_tot = 0._dl
        EFT_gpinu_tot  = 0._dl

        if (CP%Num_Nu_Massive /= 0) then                                            
            do nu_i = 1, CP%Nu_mass_eigenstates
                EFT_grhonu    = 0._dl
                EFT_gpinu     = 0._dl
                grhormass_t=grhormass(nu_i)/a**2
                call Nu_background(a*nu_masses(nu_i),EFT_grhonu,EFT_gpinu)
                EFT_grhonu_tot = EFT_grhonu_tot + grhormass_t*EFT_grhonu
                EFT_gpinu_tot  = EFT_gpinu_tot + grhormass_t*EFT_gpinu
            end do
        end if

        grho  = grho  + EFT_grhonu_tot
        gpres = gpres + EFT_gpinu_tot
        ! 3) Hubble and derivatives:
        adotoa=sqrt(grho/3)
        adotdota=(adotoa*adotoa-gpres)/2._dl
        Hdot =adotdota-adotoa**2._dl

        EFT_gpinu_tot = 0._dl
        EFT_grhonu_tot = 0._dl
        EFT_gpinudot_tot = 0._dl

        if (CP%Num_Nu_Massive /= 0) then                                            
            do nu_i = 1, CP%Nu_mass_eigenstates
                EFT_grhonu    = 0._dl
                EFT_gpinu     = 0._dl
                grhormass_t=grhormass(nu_i)/a**2
                call Nu_background(a*nu_masses(nu_i),EFT_grhonu,EFT_gpinu)
                EFT_grhonu_tot = EFT_grhonu_tot + grhormass_t*EFT_grhonu
                EFT_gpinu_tot  = EFT_gpinu_tot + grhormass_t*EFT_gpinu
                EFT_gpinudot_tot  = EFT_gpinudot_tot + grhormass_t*(Nu_pidot(a*nu_masses(nu_i),adotoa, EFT_gpinu)&
                    & -4._dl*adotoa*EFT_gpinu)
            end do
        end if

        Hdotdot = 2._dl*adotoa*Hdot &
            & + 0.5_dl*adotoa*(grhob_t + grhoc_t + 8._dl*(grhog_t+grhor_t)/3._dl)&
            & + 0.5_dl*adotoa*grhov_t*((1._dl+EFTw(a,0))*(1._dl+3._dl*EFTw(a,0))-a*EFTw(a,1))&
            & + adotoa/6._dl*EFT_grhonu_tot -0.5_dl*adotoa*EFT_gpinu_tot -0.5_dl*EFT_gpinudot_tot   

        ! 4) EFT functions:
        EFTOmegaV   = EFTOmega(a,0)
        EFTOmegaP   = EFTOmega(a,1)
        EFTOmegaPP  = EFTOmega(a,2)
        EFTOmegaPPP = EFTOmega(a,3)
        EFTAlpha1V  = EFTAlpha1(a,0)
        EFTAlpha1P  = EFTAlpha1(a,1)
        EFTAlpha2V  = EFTAlpha2(a,0)
        EFTAlpha2P  = EFTAlpha2(a,1)
        EFTAlpha3V  = EFTAlpha3(a,0)
        EFTAlpha3P  = EFTAlpha3(a,1)
        EFTAlpha4V  = EFTAlpha4(a,0)
        EFTAlpha4P  = EFTAlpha4(a,1)
        EFTAlpha5V  = EFTAlpha5(a,0)
        EFTAlpha5P  = EFTAlpha5(a,1)
        EFTAlpha6V  = EFTAlpha6(a,0)
        EFTAlpha6P  = EFTAlpha6(a,1)

        !8*pi*G*c*a^2
        EFTc = (adotoa*adotoa - Hdot)*(EFTOmegaV + a*EFTOmegaP*0.5_dl) &
            & - 0.5_dl*a2*adotoa*adotoa*EFTOmegaPP&
            & + 0.5_dl*grhov_t*(1._dl+EFTw(a,0))
        !8*pi*G*Lambda*a^2
        EFTLambda = +EFTw(a,0)*grhov_t &
            &-EFTOmegaV*(2._dl*Hdot+adotoa**2._dl) &
            &-a*EFTOmegaP*(2._dl*adotoa**2._dl + Hdot) &
            &-a2*adotoa**2._dl*EFTOmegaPP
        !EFT C DOT: 8*pi*G*cdot*a^2
        EFTcdot = -EFTOmegaV*(Hdotdot-4._dl*adotoa*Hdot+2._dl*adotoa*adotoa*adotoa) &
            & + 0.5_dl*a*EFTOmegaP*(-Hdotdot+adotoa*Hdot+adotoa*adotoa*adotoa)&
            & +0.5_dl*a2*adotoa*EFTOmegaPP*(adotoa*adotoa-3._dl*Hdot)&
            & -0.5_dl*a*a2*adotoa*adotoa*adotoa*EFTOmegaPPP&
            & +0.5_dl*adotoa*grhov_t*(-3._dl*(1._dl+EFTw(a,0))**2 + a*EFTw(a,1))
        !EFT LAMBDA DOT: 8*pi*G*Ldot*a^2
        EFTLambdadot = -2._dl*EFTOmegaV*(Hdotdot-adotoa*Hdot-adotoa*adotoa*adotoa)&
            & - a*EFTOmegaP*(4._dl*adotoa*Hdot+Hdotdot)&
            & -a2*EFTOmegaPP*adotoa*(3._dl*Hdot+2._dl*adotoa*adotoa)&
            & -a*a2*EFTOmegaPPP*adotoa*adotoa*adotoa&
            & +grhov_t*adotoa*(a*EFTw(a,1)-3._dl*EFTw(a,0)*(1._dl+EFTw(a,0)))

        if (CP%EFTflag==2.and.CP%DesignerEFTmodel/=2) then
            EFTc = EFTcTemp(a,0)
            EFTcdot = EFTcTemp(a,1)
            EFTLambda = EFTLambdaTemp(a,0)
            EFTLambdadot = EFTLambdaTemp(a,1)
        end if

        ! 5) Effective EFT quantities:
        EFTgrhoq = 2._dl*EFTc -EFTLambda -3._dl*a*adotoa*adotoa*EFTOmegaP
        EFTgpresq = EFTLambda + a2*adotoa*adotoa*EFTOmegaPP&
            & +a*EFTOmegaP*(Hdot + 2._dl*adotoa*adotoa)
        EFTgrhodotq = -3._dl*adotoa*(EFTgrhoq+EFTgpresq) + 3._dl*a*adotoa**3._dl*EFTOmegaP
        EFTgpresdotq = EFTLambdadot +a2*a*adotoa**3*EFTOmegaPPP + 3._dl*a2*adotoa*Hdot*EFTOmegaPP&
            & +a*EFTOmegaP*Hdotdot +3._dl*a*adotoa*Hdot*EFTOmegaP +2._dl*a2*adotoa**3*EFTOmegaPP&
            & -2._dl*a*adotoa**3*EFTOmegaP
        ! 6) Pi field equation coefficients:
        ! We can compute stability just for k=0 (see the notes)
        k = 0._dl
        k2 = 0._dl

        ! Coefficient A of tensors:
        EFTAT = 1._dl + EFTOmegaV + EFTAlpha4V**2
        ! First part of the A coefficient:
        EFTpiA1 = EFTc +2*a2*EFT_H0**2*EFTAlpha1V**4 +3._dl/2._dl*a2*(adotoa*EFTOmegaP+EFT_H0*EFTAlpha2V**3)**2&
            &/(2._dl*(1+EFTOmegaV)+EFTAlpha3V**2+EFTAlpha4V**2)
        !
        EFTpiA2 = +4._dl*a2*EFTAlpha6V**2
        !
        EFTpiB = EFTcdot +4._dl*adotoa*EFTc +8._dl*a2*adotoa*EFT_H0*(EFTAlpha1V**4+a*EFTAlpha1V**3*EFTAlpha1P)&
            & +4._dl*a2*k2*adotoa*(3._dl*EFTAlpha6V**2+a*EFTAlpha6V*EFTAlpha6P) +a*k2*(EFTAlpha4V**2&
            & +2._dl*EFTAlpha5V**2)/(2._dl*(1._dl+EFTOmegaV)-2._dl*EFTAlpha4V**2)*(adotoa*EFTOmegaP+EFT_H0*EFTAlpha2V**3)&
            & -a*(adotoa*EFTOmegaP+EFT_H0*EFTAlpha2V**3)/(4._dl*(1._dl+EFTOmegaV)+6._dl*EFTAlpha3V**2 +2._dl*EFTAlpha4V**2)*&
            &(-3._dl*(EFTgrhoQ + EFTgpresQ) -3._dl*a*adotoa**2*EFTOmegaP*(4._dl +Hdot/(adotoa**2)) -3._dl*a2*adotoa**2*EFTOmegaPP&
            & -3*a*adotoa*EFT_H0*(4._dl*EFTAlpha2V**3 +3._dl*a*EFTAlpha2V**2*EFTAlpha2P) -(9._dl*EFTAlpha3V**2-3._dl*EFTAlpha4V**2)*&
            &(Hdot-adotoa**2) +k2*(3._dl*EFTAlpha3V**2-EFTAlpha4V**2+4._dl*EFTAlpha5V**2))&
            & +1._dl/(1._dl+EFTOmegaV+2._dl*EFTAlpha5V**2)*(a*adotoa*EFTOmegaP+2._dl*adotoa*(EFTAlpha5V**2+2._dl*EFTAlpha5V*EFTAlpha5P)&
            & -(1._dl+EFTOmegaV)*(a*adotoa*EFTOmegaP+a*EFT_H0*EFTAlpha2V**3)/(2._dl*(1._dl+EFTOmegaV)+3._dl*EFTAlpha3V**2+EFTAlpha4V**2))*&
            &(-EFTc +1.5_dl*a*adotoa**2*EFTOmegaP-2._dl*a2*EFT_H0*EFTAlpha1V**4-4._dl*EFTAlpha6V**2*k2+1.5_dl*a*adotoa*EFT_H0*EFTAlpha2V**3)
        !
        EFTpiC = +adotoa*EFTcdot +(6._dl*adotoa**2-2._dl*Hdot)*EFTc +1.5_dl*a*adotoa*EFTOmegaP*(Hdotdot-2._dl*adotoa**3) &
            & +6._dl*a2*adotoa**2*EFT_H0**2*EFTAlpha1V**4 +2._dl*a2*Hdot*EFT_H0**2*EFTAlpha1V**4 &
            & +8._dl*a2*a*adotoa**2*EFT_H0**2*EFTAlpha1V**3*EFTAlpha1P +1.5_dl*(Hdot-adotoa**2)**2*(EFTAlpha4V**2+3._dl*EFTAlpha3V**2)&
            & +4.5_dl*adotoa*EFT_H0*a*(Hdot-adotoa**2)*(EFTAlpha2V**3+a*EFTAlpha2V**2*EFTAlpha2P)&
            & +0.5_dl*a*EFT_H0*EFTAlpha2V**3*(3._dl*Hdotdot -12._dl*Hdot*adotoa +6._dl*adotoa**3) &
            & -a*(adotoa*EFTOmegaP+EFT_H0*EFTAlpha2V**3)/(4._dl*(1._dl+EFTOmegaV)+6._dl*EFTAlpha3V**2 +2._dl*EFTAlpha4V**2)*&
            &(-3._dl*EFTgpresdotQ -3._dl*adotoa*(EFTgrhoQ +EFTgpresQ)-3._dl*a*adotoa**3*(a*EFTOmegaPP +6._dl*EFTOmegaP) &
            & -6._dl*a*adotoa*Hdot*EFTOmegaP +3._dl*(Hdotdot-2._dl*adotoa*Hdot)*(EFTAlpha4V**2+3._dl*EFTAlpha3V**2)&
            & +6._dl*adotoa*(Hdot-adotoa**2)*(3._dl*EFTAlpha3V**2+3._dl*a*EFTAlpha3V*EFTAlpha3P +EFTAlpha4V**2 +a*EFTAlpha4V*EFTAlpha4P)&
            & -3._dl*a*EFT_H0*(3._dl*adotoa**2*EFTAlpha2V**3 +Hdot*EFTAlpha2V**3 +3._dl*a*adotoa**2*EFTAlpha2V**2*EFTAlpha2P))&
            & +1._dl/(1._dl+EFTOmegaV+2._dl*EFTAlpha5V**2)*(a*adotoa*EFTOmegaP+2._dl*adotoa*(EFTAlpha5V**2+2._dl*a*EFTAlpha5V*EFTAlpha5P)&
            & -(1._dl+EFTOmegaV)*(a*adotoa*EFTOmegaP+a*EFT_H0*EFTAlpha2V**3)/(2._dl*(1._dl+EFTOmegaV)+3._dl*EFTAlpha3V**2+EFTAlpha4V**2))*&
            &(-0.5*EFTgrhodotQ -adotoa*EFTc +1.5_dl*a*adotoa*EFTOmegaP*(3._dl*adotoa**2-Hdot) -2._dl*a2*adotoa*EFT_H0**2*EFTAlpha1V**4&
            & -1.5_dl*a*EFT_H0*EFTAlpha2V**3*(Hdot-2._dl*adotoa**2) -3._dl*adotoa*(Hdot-adotoa**2)*(1.5_dl*EFTAlpha3V**2+EFTAlpha4V**2))
        ! Scale independent part of D
        EFTpiD1 = EFTc -0.5_dl*a*adotoa*EFT_H0*(EFTAlpha2V**3+3._dl*a*EFTAlpha2V**2*EFTAlpha2P) +(adotoa**2-Hdot)*(3._dl*EFTAlpha3V**2+EFTAlpha4V**2)&
            & +4._dl*a2*(Hdot*EFTAlpha6V**2 +2._dl*adotoa**2*EFTAlpha6V**2 +a*adotoa**2*EFTAlpha6V*EFTAlpha6P)&
            & +2._dl*(Hdot*EFTAlpha5V**2 +2._dl*a*adotoa**2*EFTAlpha5V*EFTAlpha5P)&
            & -a*(adotoa*EFTOmegaP+EFT_H0*EFTAlpha2V**3)/(4._dl*(1._dl+EFTOmegaV)+6._dl*EFTAlpha3V**2 +2._dl*EFTAlpha4V**2)*&
            &(-2._dl*a*adotoa*EFTOmegaP +4._dl*adotoa*EFTAlpha5V**2 -2._dl*adotoa*(3._dl*EFTAlpha3V**2 +3._dl*a*EFTAlpha3V*EFTAlpha3P &
            & +EFTAlpha4V**2 +a*EFTAlpha4V*EFTAlpha4P))&
            & +1._dl/(1._dl+EFTOmegaV+2._dl*EFTAlpha5V**2)*(a*adotoa*EFTOmegaP+2._dl*adotoa*(EFTAlpha5V**2+2._dl*a*EFTAlpha5V*EFTAlpha5P)&
            & -(1._dl+EFTOmegaV)*(a*adotoa*EFTOmegaP+a*EFT_H0*EFTAlpha2V**3)/(2._dl*(1._dl+EFTOmegaV)+3._dl*EFTAlpha3V**2+EFTAlpha4V**2))*&
            &(+0.5_dl*a*adotoa*EFTOmegaP -2._dl*adotoa*EFTAlpha5V**2 +0.5_dl*a*EFT_H0*EFTAlpha2V**3 +1.5_dl*adotoa*EFTAlpha3V**2&
            & -adotoa*EFTAlpha4V**2 -4._dl*adotoa*EFTAlpha6V**2)&
            & +(EFTAlpha4V**2 +2._dl*EFTAlpha5V**2)/(2._dl*(1._dl+EFTOmegaV) -2._dl*EFTAlpha4V**2)*(EFTgrhoQ +EFTgpresQ +a*adotoa**2*EFTOmegaP&
            & -EFTAlpha4V**2*(Hdot-adotoa**2) +a*adotoa*EFT_H0*EFTAlpha2V**3 +3._dl*EFTAlpha3V**2*(adotoa**2-Hdot))
        ! Scale dependent part of D
        EFTpiD2 = +(+0.5_dl*EFTAlpha3V**2 +0.5_dl*EFTAlpha4V**2 &
            & +(EFTAlpha4V**2 +2._dl*EFTAlpha5V**2)/(2._dl*(1._dl+EFTOmegaV) -2._dl*EFTAlpha4V**2)*(EFTAlpha3V**2+EFTAlpha4V**2))

        ! 7) Stability check:
        EFTStabilityComputation = .true.

        !1- Positive gravitational constant:
        if (1._dl +EFTOmegaV <= 0) then
            EFTStabilityComputation = .false.
            if (Feedbacklevel > 0) write(*,*) 'EFTCAMB: negative gravitational constant'
        end if
        !2- Ghost instability:
        if (EFTpiA1 < 0) then
            EFTStabilityComputation = .false.
            if (Feedbacklevel > 0) write(*,*) 'EFTCAMB: ghost instability'
        end if
        !3- Gradient instability 1:
        if (EFTpiD1 < 0) then
            EFTStabilityComputation = .false.
            if (Feedbacklevel > 0) write(*,*) 'EFTCAMB: gradient instability k^2'
        end if
        !4- Gradient instability 2:
        if (EFTpiD2 < 0) then
            EFTStabilityComputation = .false.
            if (Feedbacklevel > 0) write(*,*) 'EFTCAMB: gradient instability k^4'
        end if
        !5- Positive effective mass of pi:
        if (EFTpiC < 0.and.EFTpiMassPrior) then
            EFTStabilityComputation = .false.
            if (Feedbacklevel > 0) write(*,*) 'EFTCAMB: negative mass'
        end if
        ! 6- No tensor ghosts:
        if (EFTAT < 0) then
            EFTStabilityComputation = .false.
            if (Feedbacklevel > 0) write(*,*) 'EFTCAMB: tensor ghost instability'
        end if
        ! 7- Sub-luminal propagation:
        if (EFTlightspeedPrior) then
            if (EFTpiA2==0.and.EFTpiD2==0.and.(EFTpiD1/EFTpiA1)>1.001_dl) then
                EFTStabilityComputation = .false.
                if (Feedbacklevel > 0) write(*,*) 'EFTCAMB: tachion perturbations'
            else if (EFTpiA2/=0.and.EFTpiD2/=0.and.(EFTpiD2/EFTpiA2)>1.001_dl) then
                EFTStabilityComputation = .false.
                if (Feedbacklevel > 0) write(*,*) 'EFTCAMB: tachion perturbations'
            end if
        end if

        ! 8- Every theory has it's own peculiarities...
        ! 1) F(R): for this model it is easy to show that the positive mass condition requires that OmegaPrime
        !    should be positive. We add this test for this models as it is numerically easier to check.
        if (CP%EFTflag==2.and.CP%DesignerEFTmodel==1) then
            if (EFTOmega(0.11_dl*EFTturnonpiInitial,1)*EFTOmegaP<0) EFTStabilityComputation = .false.
        end if

        ! 9- Additional priors:
        if (EFTAdditionalPriors) then
            if (EFTw(a,0)>-1._dl/3._dl) EFTStabilityComputation = .false.
        end if

        return
    end function EFTStabilityComputation

end module EFTstability

! -------------------------------------------------------------------------------------------------

!Performs EFT initialization just after CAMB is called.
module EFTinitialization
    use EFTDef
    use EFTdeEOS
    use EFTfunctions
    use EFTreturntoGR
    use EFTstability
    implicit none

contains

    ! ----------------------------

    subroutine EFTCAMB_initialization(success)
        implicit none
        logical  :: success
        real(dl) :: astart, aend

        integer :: ind

        success = .true.

        ! 1) Feedback
        if (Feedbacklevel>0) write(*,*) 'EFTCAMB initialization'
        if (Feedbacklevel>2) call EFTCAMB_print_parameters

        ! 2) EFT hard priors
        call EFTCAMB_HardPriors(success)
        if (.not.success) return

        ! 3) Call the designer
        if (CP%EFTflag==2) then
            call EFTCAMB_Designer(success)
            if (.not.success) return
        end if

        ! 4) Check stability
        if (EarlyTimeStability) then
            astart = 10._dl**(-7.8_dl)
        else
            astart = EFTturnonpiInitial
        end if

        aend   = 1._dl
        call EFTCheck_Stability(success, astart, aend)
        if (.not.success) return

        ! 5) Calculate Return to GR
        call EFTCheckReturnToGR

    end subroutine EFTCAMB_initialization

    ! ----------------------------

    subroutine EFTCAMB_HardPriors(success)
        ! This subroutine enforces hard priors on cosmological parameters to avoid EFTCAMB
        ! being called for weird values of cosmological parameters.
        implicit none
        logical  :: success
        real(dl) :: a, amin, amax
        integer  :: i

        ! 1) There should be dark energy:
        if (CP%Omegav<0.and.EFTAdditionalPriors) then
            success=.false.
            return
        end if

        ! 2) Dark energy should source cosmic acceleration:
        if (EFTAdditionalPriors) then
            do i=1, 100
                amin = 0._dl
                amax = 1._dl
                a = amin +REAL(i-1)*(amax-amin)/REAL(100-1)
                if (EFTw(a,0)>-1._dl/3._dl) then
                    success=.false.
                    return
                end if
            end do
        end if


    end subroutine EFTCAMB_HardPriors

    ! ----------------------------

    subroutine EFTCAMB_print_parameters
        ! This is used to print EFT parameters when debugging the code.
        implicit none

        write(*,*)
        write(*,*) 'EFTCAMB model flags:'
        write(*,*) 'EFTflag = ', CP%EFTflag
        write(*,*)

        if (CP%EFTflag==1.or.CP%EFTflag==2) write(*,*) 'w_DE =', CP%EFTwDE
        if (CP%EFTflag==2) write(*,*) 'DesignerEFTmodel =', CP%DesignerEFTmodel
        if (CP%EFTflag==3) write(*,*) 'MatchingEFTmodel =', CP%DesignerEFTmodel
        if (CP%EFTflag==1) then
            write(*,*) 'PureEFTmodelOmega  =', CP%PureEFTmodelOmega
            write(*,*) 'PureEFTmodelAlpha1 =', CP%PureEFTmodelAlpha1
            write(*,*) 'PureEFTmodelAlpha2 =', CP%PureEFTmodelAlpha2
            write(*,*) 'PureEFTmodelAlpha3 =', CP%PureEFTmodelAlpha3
            write(*,*) 'PureEFTmodelAlpha4 =', CP%PureEFTmodelAlpha4
            write(*,*) 'PureEFTmodelAlpha5 =', CP%PureEFTmodelAlpha5
            write(*,*) 'PureEFTmodelAlpha6 =', CP%PureEFTmodelAlpha6
        end if

        write(*,*)
        write(*,*) 'EFTCAMB parameters:'
        write(*,*)

        if ((CP%EFTflag==1.or.CP%EFTflag==2).and.CP%EFTwDE/=0) then
            write(*,*) 'EFTw0  =', CP%EFTw0
            write(*,*) 'EFTwa  =', CP%EFTwa
            write(*,*) 'EFTwn  =', CP%EFTwn
            write(*,*) 'EFTwat =', CP%EFTwat
            write(*,*) 'EFtw2  =', CP%EFtw2
            write(*,*) 'EFTw3  =', CP%EFTw3
        end if

        if (CP%EFTflag==1) then
            if (CP%PureEFTmodelOmega/=0) write(*,*) 'EFTOmega0    =', CP%EFTOmega0
            if (CP%PureEFTmodelOmega/=0) write(*,*) 'EFTOmegaExp  =', CP%EFTOmegaExp
            if (CP%PureEFTmodelAlpha1/=0) write(*,*) 'EFTAlpha10   =', CP%EFTAlpha10
            if (CP%PureEFTmodelAlpha1/=0) write(*,*) 'EFTAlpha1Exp =', CP%EFTAlpha1Exp
            if (CP%PureEFTmodelAlpha2/=0) write(*,*) 'EFTAlpha20   =', CP%EFTAlpha20
            if (CP%PureEFTmodelAlpha2/=0) write(*,*) 'EFTAlpha2Exp =', CP%EFTAlpha2Exp
            if (CP%PureEFTmodelAlpha3/=0) write(*,*) 'EFTAlpha30   =', CP%EFTAlpha30
            if (CP%PureEFTmodelAlpha3/=0) write(*,*) 'EFTAlpha3Exp =', CP%EFTAlpha3Exp
            if (CP%PureEFTmodelAlpha4/=0) write(*,*) 'EFTAlpha40   =', CP%EFTAlpha40
            if (CP%PureEFTmodelAlpha4/=0) write(*,*) 'EFTAlpha4Exp =', CP%EFTAlpha4Exp
            if (CP%PureEFTmodelAlpha5/=0) write(*,*) 'EFTAlpha50   =', CP%EFTAlpha50
            if (CP%PureEFTmodelAlpha5/=0) write(*,*) 'EFTAlpha5Exp =', CP%EFTAlpha5Exp
            if (CP%PureEFTmodelAlpha6/=0) write(*,*) 'EFTAlpha60   =', CP%EFTAlpha50
            if (CP%PureEFTmodelAlpha6/=0) write(*,*) 'EFTAlpha6Exp =', CP%EFTAlpha5Exp
        end if

        if (CP%EFTflag==2) then
            if (CP%DesignerEFTmodel==1) write(*,*) 'F(R) B0 =',  CP%EFTB0
        end if

        write(*,*)

    end subroutine EFTCAMB_print_parameters

end module EFTinitialization

! -------------------------------------------------------------------------------------------------
