! EFTCAMB: Simple program that will sample parameter space to output the result of the stability
!          of the dark energy / Modified Gravity model considered.
!
! This program requires compiling with camb but actually does not use the parameter file.
!

! -------------------------------------------------------------------------------------------------

module StabilityFunctions
    use precision
    use CAMB
    implicit none

    type(CAMBparams), target :: P
    real(dl), pointer :: EFTparam
    real(dl), pointer :: EFTparam2
    real(dl), pointer :: EFTparam3

contains

    ! ----------------------------

    function Test1D(param)
        ! Multiple step function for 1D test purposes.
        implicit none

        real(dl) :: Test1D
        real(dl), intent(in) :: param

        if (param>0._dl) then
            Test1D = 1._dl
        else
            Test1D = 0._dl
        end if

        return
    end function Test1D

    ! ----------------------------

    function Test2D(param1, param2)
        ! Binary function for 2D test purposes.
        implicit none

        real(dl) :: Test2D
        real(dl), intent(in) :: param1, param2

        if (param1 > 0._dl.and.param2 > 0._dl) then
            Test2D = 1._dl
        else
            Test2D = 0._dl
        end if

        return
    end function Test2D

    ! ----------------------------

    function Test3D(param1, param2, param3)
        ! Binary function for 3D test purposes.
        implicit none

        real(dl) :: Test3D
        real(dl), intent(in) :: param1, param2, param3

        if (param1 > 0._dl.and.param2 > 0._dl.and.param3 > 0._dl) then
            Test3D = 1._dl
        else
            Test3D = 0._dl
        end if

        return
    end function Test3D

    ! ----------------------------

    function Stability1D(param)
        ! Binary function that gives the stability varying 1 parameter.
        use EFTinitialization
        implicit none

        real(dl) :: Stability1D
        real(dl), intent(in) :: param
        logical  :: success

        ! 1) Set the cosmological parameter that is varied:
        EFTparam = param
        ! 2) Call the CAMB subroutine to set cosmological parameters:
        call CAMBParams_Set(P)
        success = .true.
        ! 3) Call the EFT initialization module to check the stability:
        call EFTCAMB_initialization(success)
        ! 4) Write the result:
        if (success) then
            Stability1D = 1._dl
        else
            Stability1D = 0._dl
        end if

        return
    end function Stability1D

    ! ----------------------------

    function Stability2D(param1, param2)
        ! Binary function that gives the stability varying 2 parameters.
        use EFTinitialization
        implicit none

        real(dl) :: Stability2D
        real(dl), intent(in) :: param1, param2
        logical  :: success

        ! 1) Set the cosmological parameter that is varied:
        EFTparam  = param1
        EFTparam2 = param2
        ! 2) Call the CAMB subroutine to set cosmological parameters:
        call CAMBParams_Set(P)
        success = .true.
        ! 3) Call the EFT initialization module to check the stability:
        call EFTCAMB_initialization(success)
        ! 4) Write the result:
        if (success) then
            Stability2D = 1._dl
        else
            Stability2D = 0._dl
        end if

        return
    end function Stability2D

     ! ----------------------------

    function Stability3D(param1, param2, param3)
        ! Binary function that gives the stability varying 2 parameters.
        use EFTinitialization
        implicit none

        real(dl) :: Stability3D
        real(dl), intent(in) :: param1, param2, param3
        logical  :: success

        ! 1) Set the cosmological parameter that is varied:
        EFTparam  = param1
        EFTparam2 = param2
        EFTparam3 = param3
        ! 2) Call the CAMB subroutine to set cosmological parameters:
        call CAMBParams_Set(P)
        success = .true.
        ! 3) Call the EFT initialization module to check the stability:
        call EFTCAMB_initialization(success)
        ! 4) Write the result:
        if (success) then
            Stability3D = 1._dl
        else
            Stability3D = 0._dl
        end if

        return
    end function Stability3D

end module StabilityFunctions

! -------------------------------------------------------------------------------------------------

module SamplersSubroutines
    use precision
    implicit none
contains

    ! ----------------------------

    subroutine Grid_Sampling1D(function, IntMin, IntMax, Points, Results)
        ! Subroutine that performs sampling of a 1D function on a grid of points.
        !
        ! function = function to be sampled                                     (in)
        ! IntMin   = minimum of the interval in which the function is sampled   (in)
        ! IntMax   = maximum of the interval in which the function is sampled   (in)
        ! Points   = number of points to sample                                 (in)
        ! Results  = input: an allocatable array, output: allocated array of    (in out)
        !            dimension (2,Points) containing the points of the form
        !            (x,function(x))
        !
        implicit none

        interface
            function function(x)
                use precision
                real(dl) ::  function
                real(dl), intent(in) :: x
            end function function
        end interface

        real(dl), intent(in)  :: IntMin, IntMax
        integer , intent(in)  :: Points
        real(dl), allocatable :: Results(:,:)

        integer  :: i,j
        real(dl) :: x, y

        ! 1) Deallocate and allocate the results array.
        if (allocated(Results)) deallocate(Results)
        allocate(Results(2,Points))
        ! 2) Do the sampling
        do i = 1, Points
            x = IntMin +REAL(i-1)*(IntMax-IntMin)/REAL(Points-1)
            y = function(x)

            Results(1,i) = x
            Results(2,i) = y
        end do

        return
    end subroutine Grid_Sampling1D

    ! ----------------------------

    subroutine Adaptive_Sampling1D(function, IntMin, IntMax, Points, Results, Recursions)
        ! Subroutine that performs an adaptively refined sampling of a 1D function.
        ! Much more efficient than grid sampling.
        !
        ! function   = function to be sampled                                     (in)
        ! IntMin     = minimum of the interval in which the function is sampled   (in)
        ! IntMax     = maximum of the interval in which the function is sampled   (in)
        ! Points     = number of points of the initial grid sampling              (in)
        ! Results    = input: an allocatable array, output: allocated array       (in out)
        !              containing the points of the form (x,function(x))
        ! Recursions = (OPTIONAL) maximum number of points in which               (in)
        !              every sub-interval is adaptively sampled.
        !
        implicit none

        interface
            function function(x)
                use precision
                real(dl) ::  function
                real(dl), intent(in) :: x
            end function function
        end interface

        real(dl), intent(in)  :: IntMin, IntMax
        integer , intent(in)  :: Points
        integer , optional :: Recursions

        real(dl), allocatable :: Results(:,:)
        real(dl), allocatable :: Temp(:,:), Temp2(:,:)

        integer  :: i,j
        integer  :: GlobalCounter, LocalCounter
        real(dl) :: x, y, x1, x2, y1, y2

        ! 1) If the maximum number of recursions is not in input set it to 5
        if (.not.present(Recursions)) Recursions = 5
        ! 2) Allocate Temp as the maximum number of points possible given Recursions
        if (allocated(Temp))  deallocate(Temp)
        if (allocated(Temp2)) deallocate(Temp2)
        allocate(Temp(2,Points*(Recursions+1)),Temp2(2,Points))
        ! 3) Do the sampling
        GlobalCounter = 1
        do i = 1, Points
            x = IntMin +REAL(i-1)*(IntMax-IntMin)/REAL(Points-1)
            y = function(x)
            Temp2(1,i) = x
            Temp2(2,i) = y
            if (i>1) then
                LocalCounter  = 0
                x1 = Temp2(1,i-1)
                x2 = Temp2(1,i)
                y1 = Temp2(2,i-1)
                y2 = Temp2(2,i)
                call Interval_Recursive_Sampler(function, x1, x2, y1, y2, LocalCounter, Recursions,&
                    & Temp, Points*(Recursions+1), GlobalCounter)
            end if
            Temp(1,GlobalCounter) = x
            Temp(2,GlobalCounter) = y
            GlobalCounter = GlobalCounter + 1
        end do

        ! 4) Write out results
        if (allocated(Results)) deallocate(Results)
        allocate(Results(2,GlobalCounter-1))
        do i = 1, GlobalCounter-1
            Results(1,i) = Temp(1,i)
            Results(2,i) = Temp(2,i)
        end do

        ! 5) Free memory
        if (allocated(Temp))  deallocate(Temp)
        if (allocated(Temp2)) deallocate(Temp2)

        return
    end subroutine Adaptive_Sampling1D

    ! ----------------------------

    recursive subroutine Interval_Recursive_Sampler(function, x1, x2, y1, y2, counter, MaxRecursion,&
        & ResArray, ResArrayLength, ResArrayInd)
        ! Subdivides an intervall and decide wether to sample it recursively.
        ! This works good for our purposes but not so well for a generic function.
        ! The criterium for refining is based on a combination of first and second derivatives.
        implicit none

        external function
        real(dl) :: function

        real(dl), intent(inout) :: x1, x2, y1, y2
        integer , intent(inout) :: counter
        integer , intent(in)    :: MaxRecursion
        integer  :: ResArrayLength, ResArrayInd
        real(dl) :: ResArray(2,ResArrayLength)


        real(dl), parameter :: tol = 0.1_dl
        real(dl), parameter :: cut = 1000._dl

        real(dl) :: xm, ym, ep
        real(dl) :: der2

        if (counter >= MaxRecursion) return

        ep = ABS(y2-y1)/ABS(x2-x1)

        if (ep>tol.and.ep<cut) then
            xm = 0.5_dl*(x1+x2)
            ym = function(xm)
            der2 = (y1 -2._dl*ym +y2)/(x2-xm)**2

            counter = counter + 1

            ResArray(1,ResArrayInd) = xm
            ResArray(2,ResArrayInd) = ym

            ResArrayInd = ResArrayInd +1

            if (ABS(der2)>0.1_dl*tol) then
                call Interval_Recursive_Sampler(function, x1, xm, y1, ym, counter, MaxRecursion,&
                    &ResArray, ResArrayLength, ResArrayInd)
                call Interval_Recursive_Sampler(function, xm, x2, ym, y2, counter, MaxRecursion,&
                    &ResArray, ResArrayLength, ResArrayInd)
            end if
        else
            return
        end if

    end subroutine Interval_Recursive_Sampler

    ! ----------------------------

    subroutine Grid_Sampling2D(function, IntMin1, IntMax1, IntMin2, IntMax2, Points, Results)
        ! function = function to be sampled                                     (in)
        ! IntMin1  = minimum of the interval in which the function is sampled   (in)
        !            in the first parameter.
        ! IntMax1  = maximum of the interval in which the function is sampled   (in)
        !            in the first parameter.
        ! IntMin2  = minimum of the interval in which the function is sampled   (in)
        !            in the second parameter.
        ! IntMax2  = maximum of the interval in which the function is sampled   (in)
        !            in the second parameter.
        ! Points   = number of points in which both intervals are subdivided    (in)
        !            the total number of sampled points would be Points**2
        ! Results  = input: an allocatable array, output: allocated array of    (in out)
        !            dimension (2,Points) containing the points of the form
        !            (x,function(x))
        implicit none
        ! MRTODO: write the interface for f
        external function
        real(dl) :: function

        real(dl), intent(in)  :: IntMin1, IntMax1, IntMin2, IntMax2
        integer , intent(in)  :: Points
        real(dl), allocatable :: Results(:,:)

        integer  :: i,j,k
        real(dl) :: x, y, z

        ! 1) Deallocate and allocate the results array.
        if (allocated(Results)) deallocate(Results)
        allocate(Results(3,Points**2))
        ! 2) Do the sampling
        k=0
        do i = 1, Points
            do j = 1, Points

                x = IntMin1 +REAL(i-1)*(IntMax1-IntMin1)/REAL(Points-1)
                y = IntMin2 +REAL(j-1)*(IntMax2-IntMin2)/REAL(Points-1)
                z = function(x,y)

                k = k +1

                Results(1,k) = x
                Results(2,k) = y
                Results(3,k) = z

            end do
        end do

        return
    end subroutine Grid_Sampling2D

    ! ----------------------------

    subroutine Grid_Sampling3D(function, IntMin1, IntMax1, IntMin2, IntMax2, IntMin3, IntMax3, Points, Results)
        ! function = function to be sampled                                     (in)
        ! IntMin1  = minimum of the interval in which the function is sampled   (in)
        !            in the first parameter.
        ! IntMax1  = maximum of the interval in which the function is sampled   (in)
        !            in the first parameter.
        ! IntMin2  = minimum of the interval in which the function is sampled   (in)
        !            in the second parameter.
        ! IntMax2  = maximum of the interval in which the function is sampled   (in)
        !            in the second parameter.
        ! IntMin3  = minimum of the interval in which the function is sampled   (in)
        !            in the third parameter.
        ! IntMax3  = maximum of the interval in which the function is sampled   (in)
        !            in the third parameter.
        ! Points   = number of points in which both intervals are subdivided    (in)
        !            the total number of sampled points would be Points**2
        ! Results  = input: an allocatable array, output: allocated array of    (in out)
        !            dimension (2,Points) containing the points of the form
        !            (x,function(x))
        implicit none
        ! MRTODO: write the interface for f
        external function
        real(dl) :: function

        real(dl), intent(in)  :: IntMin1, IntMax1, IntMin2, IntMax2, IntMin3, IntMax3
        integer , intent(in)  :: Points
        real(dl), allocatable :: Results(:,:)

        integer  :: i,j,k,m
        real(dl) :: x, y, z, f

        ! 1) Deallocate and allocate the results array.
        if (allocated(Results)) deallocate(Results)
        allocate(Results(4,Points**3))
        ! 2) Do the sampling
        k=0
        do i = 1, Points
            do j = 1, Points
                do m = 1, Points

                    x = IntMin1 +REAL(i-1)*(IntMax1-IntMin1)/REAL(Points-1)
                    y = IntMin2 +REAL(j-1)*(IntMax2-IntMin2)/REAL(Points-1)
                    z = IntMin3 +REAL(m-1)*(IntMax3-IntMin3)/REAL(Points-1)
                    f = function(x,y,z)

                    k = k +1

                    Results(1,k) = x
                    Results(2,k) = y
                    Results(3,k) = z
                    Results(4,k) = f

                end do
            end do
        end do

        return
    end subroutine Grid_Sampling3D

end module SamplersSubroutines

! -------------------------------------------------------------------------------------------------

module Systematic_Model_Sampling
    use CAMB
    use EFTinitialization
    use StabilityFunctions
    use SamplersSubroutines
    implicit none

contains

    ! ----------------------------

    subroutine EFT_Clean_Params(P)
        ! Subroutines that resets the values of the EFT parameters
        implicit none
        type(CAMBparams) :: P
        ! 1) Clear flags:
        P%EFTflag = 0
        P%EFTwDE  = 0
        P%PureEFTmodelOmega  = 0
        P%PureEFTmodelAlpha1 = 0
        P%PureEFTmodelAlpha2 = 0
        P%PureEFTmodelAlpha3 = 0
        P%PureEFTmodelAlpha4 = 0
        P%PureEFTmodelAlpha5 = 0
        P%PureEFTmodelAlpha6 = 0
        P%DesignerEFTmodel   = 0
        P%MatchingEFTmodel   = 0
        ! 2) Clear model parameters:
        P%EFTw0 = 0._dl
        P%EFTwa = 0._dl
        P%EFTwn = 0._dl
        P%EFTwat= 0._dl
        P%EFtw2 = 0._dl
        P%EFTw3 = 0._dl
        P%EFTOmega0    = 0._dl
        P%EFTOmegaExp  = 0._dl
        P%EFTAlpha10   = 0._dl
        P%EFTAlpha1Exp = 0._dl
        P%EFTAlpha20   = 0._dl
        P%EFTAlpha2Exp = 0._dl
        P%EFTAlpha30   = 0._dl
        P%EFTAlpha3Exp = 0._dl
        P%EFTAlpha40   = 0._dl
        P%EFTAlpha4Exp = 0._dl
        P%EFTAlpha50   = 0._dl
        P%EFTAlpha5Exp = 0._dl
        P%EFTAlpha60   = 0._dl
        P%EFTAlpha6Exp = 0._dl
        P%EFTB0 = 0._dl
    end subroutine EFT_Clean_Params

    ! ----------------------------

    subroutine SamplePureEFT1D(NumPoints, IntMin, IntMax, MaxRecursions)
        ! Subroutine to test at once every 1 parameter, pure EFT, extension of the standard model.
        implicit none

        integer , intent(in) :: NumPoints
        real(dl), intent(in) :: IntMin, IntMax
        integer, optional :: MaxRecursions

        character(100) :: OutDir
        real(dl), allocatable :: res(:,:)

        ! 1) Set MaxRecursions
        if (.not.present(MaxRecursions)) MaxRecursions = 5
        ! 2) Set the output directory
        OutDir = 'Results/Stability/1_PureEFT_1D/'
        ! 3) Some feedback
        write(*,*) 'Starting Pure EFT stability check:'
        write(*,*) 'results files placed in: '//TRIM(OutDir)

        ! 1) Pure EFT Omega, constant
        call EFT_Clean_Params(P)
        P%EFTflag = 1
        P%PureEFTmodelOmega  = 1
        EFTparam => P%EFTOmega0
        call Adaptive_Sampling1D(Stability1D, IntMin, IntMax, NumPoints, res, Recursions = MaxRecursions)
        call CreateTxtFile(TRIM(OutDir)//'/1_Omega_1D_1.dat',100)
        write (100,'(E15.5,E15.5)') res
        close(100)
        write(*,*) 'Done Omega 1'

        ! 2) Pure EFT Omega, linear
        call EFT_Clean_Params(P)
        P%EFTflag = 1
        P%PureEFTmodelOmega  = 2
        EFTparam => P%EFTOmega0
        call Adaptive_Sampling1D(Stability1D, IntMin, IntMax, NumPoints, res, Recursions = MaxRecursions)
        call CreateTxtFile(TRIM(OutDir)//'/1_Omega_1D_2.dat',100)
        write (100,'(E15.5,E15.5)') res
        close(100)
        write(*,*) 'Done Omega 2'

        ! 3) Pure EFT Alpha1, constant
        call EFT_Clean_Params(P)
        P%EFTflag = 1
        P%PureEFTmodelAlpha1 = 1
        EFTparam => P%EFTAlpha10
        call Adaptive_Sampling1D(Stability1D, IntMin, IntMax, NumPoints, res, Recursions = MaxRecursions)
        call CreateTxtFile(TRIM(OutDir)//'/1_Alpha1_1D_1.dat',100)
        write (100,'(E15.5,E15.5)') res
        close(100)
        write(*,*) 'Done Alpha1 1'

         ! 4) Pure EFT Alpha1, linear
        call EFT_Clean_Params(P)
        P%EFTflag = 1
        P%PureEFTmodelAlpha1 = 2
        EFTparam => P%EFTAlpha10
        call Adaptive_Sampling1D(Stability1D, IntMin, IntMax, NumPoints, res, Recursions = MaxRecursions)
        call CreateTxtFile(TRIM(OutDir)//'/1_Alpha1_1D_2.dat',100)
        write (100,'(E15.5,E15.5)') res
        close(100)
        write(*,*) 'Done Alpha1 2'

        ! 5) Pure EFT Alpha2, constant
        call EFT_Clean_Params(P)
        P%EFTflag = 1
        P%PureEFTmodelAlpha2 = 1
        EFTparam => P%EFTAlpha20
        call Adaptive_Sampling1D(Stability1D, IntMin, IntMax, NumPoints, res, Recursions = MaxRecursions)
        call CreateTxtFile(TRIM(OutDir)//'/1_Alpha2_1D_1.dat',100)
        write (100,'(E15.5,E15.5)') res
        close(100)
        write(*,*) 'Done Alpha2 1'

         ! 6) Pure EFT Alpha2, linear
        call EFT_Clean_Params(P)
        P%EFTflag = 1
        P%PureEFTmodelAlpha2 = 2
        EFTparam => P%EFTAlpha20
        call Adaptive_Sampling1D(Stability1D, IntMin, IntMax, NumPoints, res, Recursions = MaxRecursions)
        call CreateTxtFile(TRIM(OutDir)//'/1_Alpha2_1D_2.dat',100)
        write (100,'(E15.5,E15.5)') res
        close(100)
        write(*,*) 'Done Alpha2 2'

        ! 7) Pure EFT Alpha3, constant
        call EFT_Clean_Params(P)
        P%EFTflag = 1
        P%PureEFTmodelAlpha3 = 1
        EFTparam => P%EFTAlpha30
        call Adaptive_Sampling1D(Stability1D, IntMin, IntMax, NumPoints, res, Recursions = MaxRecursions)
        call CreateTxtFile(TRIM(OutDir)//'/1_Alpha3_1D_1.dat',100)
        write (100,'(E15.5,E15.5)') res
        close(100)
        write(*,*) 'Done Alpha3 1'

         ! 8) Pure EFT Alpha3, linear
        call EFT_Clean_Params(P)
        P%EFTflag = 1
        P%PureEFTmodelAlpha3 = 2
        EFTparam => P%EFTAlpha30
        call Adaptive_Sampling1D(Stability1D, IntMin, IntMax, NumPoints, res, Recursions = MaxRecursions)
        call CreateTxtFile(TRIM(OutDir)//'/1_Alpha3_1D_2.dat',100)
        write (100,'(E15.5,E15.5)') res
        close(100)
        write(*,*) 'Done Alpha3 2'

        ! 9) Pure EFT Alpha4, constant
        call EFT_Clean_Params(P)
        P%EFTflag = 1
        P%PureEFTmodelAlpha4 = 1
        EFTparam => P%EFTAlpha40
        call Adaptive_Sampling1D(Stability1D, IntMin, IntMax, NumPoints, res, Recursions = MaxRecursions)
        call CreateTxtFile(TRIM(OutDir)//'/1_Alpha4_1D_1.dat',100)
        write (100,'(E15.5,E15.5)') res
        close(100)
        write(*,*) 'Done Alpha4 1'

         ! 10) Pure EFT Alpha4, linear
        call EFT_Clean_Params(P)
        P%EFTflag = 1
        P%PureEFTmodelAlpha4 = 2
        EFTparam => P%EFTAlpha40
        call Adaptive_Sampling1D(Stability1D, IntMin, IntMax, NumPoints, res, Recursions = MaxRecursions)
        call CreateTxtFile(TRIM(OutDir)//'/1_Alpha4_1D_2.dat',100)
        write (100,'(E15.5,E15.5)') res
        close(100)
        write(*,*) 'Done Alpha4 2'

        ! 11) Pure EFT Alpha5, constant
        call EFT_Clean_Params(P)
        P%EFTflag = 1
        P%PureEFTmodelAlpha5 = 1
        EFTparam => P%EFTAlpha50
        call Adaptive_Sampling1D(Stability1D, IntMin, IntMax, NumPoints, res, Recursions = MaxRecursions)
        call CreateTxtFile(TRIM(OutDir)//'/1_Alpha5_1D_1.dat',100)
        write (100,'(E15.5,E15.5)') res
        close(100)
        write(*,*) 'Done Alpha5 1'

         ! 12) Pure EFT Alpha5, linear
        call EFT_Clean_Params(P)
        P%EFTflag = 1
        P%PureEFTmodelAlpha5 = 2
        EFTparam => P%EFTAlpha50
        call Adaptive_Sampling1D(Stability1D, IntMin, IntMax, NumPoints, res, Recursions = MaxRecursions)
        call CreateTxtFile(TRIM(OutDir)//'/1_Alpha5_1D_2.dat',100)
        write (100,'(E15.5,E15.5)') res
        close(100)
        write(*,*) 'Done Alpha5 2'

        ! 13) Pure EFT Alpha6, constant
        call EFT_Clean_Params(P)
        P%EFTflag = 1
        P%PureEFTmodelAlpha6 = 1
        EFTparam => P%EFTAlpha60
        call Adaptive_Sampling1D(Stability1D, IntMin, IntMax, NumPoints, res, Recursions = MaxRecursions)
        call CreateTxtFile(TRIM(OutDir)//'/1_Alpha6_1D_1.dat',100)
        write (100,'(E15.5,E15.5)') res
        close(100)
        write(*,*) 'Done Alpha6 1'

         ! 14) Pure EFT Alpha6, linear
        call EFT_Clean_Params(P)
        P%EFTflag = 1
        P%PureEFTmodelAlpha6 = 2
        EFTparam => P%EFTAlpha60
        call Adaptive_Sampling1D(Stability1D, IntMin, IntMax, NumPoints, res, Recursions = MaxRecursions)
        call CreateTxtFile(TRIM(OutDir)//'/1_Alpha6_1D_2.dat',100)
        write (100,'(E15.5,E15.5)') res
        close(100)
        write(*,*) 'Done Alpha6 2'

        write(*,*) 'Pure EFT stability check done.'

    end subroutine SamplePureEFT1D

end module Systematic_Model_Sampling

! -------------------------------------------------------------------------------------------------

program EFTStabilityInPSpace
    use CAMB
    use EFTinitialization
    use StabilityFunctions
    use SamplersSubroutines
    use Systematic_Model_Sampling
    implicit none

    real(dl), allocatable :: res(:,:)
    integer :: i
    integer :: x

    real(dl) :: x1,x2, y1, y2
    integer  :: counter, MaxRecursion

    ! 0) Set some useful stuff:
    feedbackLevel = 0

    ! 1) Set fixed cosmological parameters:
    P%tcmb    = 2.7255
    P%omegab  = 0.046
    P%omegac  = 0.228
    P%omegan  = 0.0
    P%H0      = 7
    P%omegav  = 1-(P%omegab+P%omegac+P%omegan)

    ! 2) Do the test:
    call SamplePureEFT1D(100, -10._dl, 10._dl, MaxRecursions = 10)

end program EFTStabilityInPSpace
